# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import MDAnalysis as mda
import pytest
from MDAnalysis.analysis.nucleicacids import WatsonCrickDist
from MDAnalysisTests.datafiles import RNA_PSF, RNA_PDB
from numpy.testing import assert_allclose


@pytest.fixture(scope='module')
def u():
    return mda.Universe(RNA_PSF, RNA_PDB)


def test_wc_dist(u):
    strand: mda.AtomGroup = u.select_atoms("segid RNAA")
    strand1 = [strand.residues[0], strand.residues[21]]
    strand2 = [strand.residues[1], strand.residues[22]]

    WC = WatsonCrickDist(strand1, strand2)
    WC.run()

    assert_allclose(WC.results[0][0], 4.3874702, atol=1e-3)
    assert_allclose(WC.results[1][0], 4.1716404, atol=1e-3)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""Tests for MDAnalysis.core.topologyattrs objects.

"""
import numpy as np

from numpy.testing import (
    assert_equal,
    assert_almost_equal,
)
import pytest
from MDAnalysisTests.datafiles import PSF, DCD, PDB_CHECK_RIGHTHAND_PA, MMTF
from MDAnalysisTests import make_Universe

import MDAnalysis as mda
import MDAnalysis.core.topologyattrs as tpattrs
from MDAnalysis.core._get_readers import get_reader_for
from MDAnalysis.core.topology import Topology
from MDAnalysis.exceptions import NoDataError


class DummyGroup(object):
    """Designed to mock an Group

    initiate with indices, these are then available as ._ix
    """
    def __init__(self, vals):
        self._ix = vals

    def __len__(self):
        return len(self._ix)

    @property
    def ix(self):
        return self._ix


class TopologyAttrMixin(object):
    """Mixin to test the common elements to all TopologyAttrs.

    10 atoms
    4 residues
    2 segments

    """
    # Reference data


    @pytest.fixture()
    def top(self):
        Ridx = np.array([0, 0, 2, 2, 1, 1, 3, 3, 1, 2])
        Sidx = np.array([0, 1, 1, 0])
        return Topology(10, 4, 2,
                        attrs=[self.attrclass(self.values.copy())],
                        atom_resindex=Ridx,
                        residue_segindex=Sidx)

    @pytest.fixture()
    def attr(self, top):
        return getattr(top, self.attrclass.attrname)

    @pytest.fixture()
    def universe(self, top):
        return mda.Universe(top)

    def test_len(self, attr):
        assert len(attr) == len(attr.values)


class TestAtomAttr(TopologyAttrMixin):
    """Test atom-level TopologyAttrs.

    """
    values = np.array([7, 3, 69, 9993, 84, 194, 263, 501, 109, 5873])
    single_value = 567
    attrclass = tpattrs.AtomAttr

    def test_set_atom_VE(self):
        u = make_Universe(('names',))
        at = u.atoms[0]

        with pytest.raises(ValueError):
            setattr(at, 'name', ['oopsy', 'daisies'])

    def test_get_atoms(self, attr):
        result = attr.get_atoms(DummyGroup([2, 1]))

        assert len(result) == 2
        assert_equal(result,
                           self.values[[2, 1]])

    def test_set_atoms_singular(self, attr):
        # set len 2 Group to len 1 value
        dg = DummyGroup([3, 7])
        attr.set_atoms(dg, self.single_value)
        assert_equal(attr.get_atoms(dg),
                     np.array([self.single_value, self.single_value]))

    def test_set_atoms_plural(self, attr):
        # set len 2 Group to len 2 values
        dg = DummyGroup([3, 7])
        attr.set_atoms(dg, np.array([23, 504]))
        assert_equal(attr.get_atoms(dg), np.array([23, 504]))

    def test_set_atoms_VE(self, attr):
        # set len 2 Group to wrong length values
        dg = DummyGroup([3, 7])
        with pytest.raises(ValueError):
            attr.set_atoms(dg, np.array([6, 7, 8, 9]))

    def test_get_residues(self, attr):
        """Unless overriden by child class, this should yield values for all
        atoms in residues.

        """
        result = attr.get_residues(DummyGroup([2, 1]))

        assert len(result) == 2
        assert_equal(result,
                           [self.values[[2, 3, 9]], self.values[[4, 5, 8]]])

    def test_get_segments(self, attr):
        """Unless overriden by child class, this should yield values for all
        atoms in segments.

        """
        result = attr.get_segments(DummyGroup([1]))

        assert len(result) == 1
        assert_equal(result,
                           [self.values[[4, 5, 8, 2, 3, 9]]])


class TestAtomids(TestAtomAttr):
    attrclass = tpattrs.Atomids


class TestIndicesClasses(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(PSF, DCD)

    def test_cant_set_atom_indices(self, u):
        with pytest.raises(AttributeError):
            u.atoms.indices = 1

    def test_cant_set_residue_indices(self, u):
        with pytest.raises(AttributeError):
            u.atoms.residues.resindices = 1

    def test_cant_set_segment_indices(self, u):
        with pytest.raises(AttributeError):
            u.atoms.segments.segindices = 1


class TestAtomnames(TestAtomAttr):
    values = np.array(['O', 'C', 'CA', 'N', 'CB', 'CG', 'CD', 'NA', 'CL', 'OW'],
                      dtype=object)
    single_value = 'Ca2'
    attrclass = tpattrs.Atomnames

    @pytest.fixture()
    def u(self):
        return mda.Universe(PSF, DCD)

    def test_prev_emptyresidue(self, u):
        # Checking group size rather than doing
        # array comparison on zero-sized arrays (Issue #3535)
        groupsize = len(u.residues[[]]._get_prev_residues_by_resid().atoms)
        assert groupsize == 0
        assert groupsize == len(u.residues[[]].atoms)

    def test_next_emptyresidue(self, u):
        # See above re: checking size for zero-sized arrays
        groupsize = len(u.residues[[]]._get_next_residues_by_resid().atoms)
        assert groupsize == 0
        assert groupsize == len(u.residues[[]].atoms)


class AggregationMixin(TestAtomAttr):
    def test_get_residues(self, attr):
        assert_equal(attr.get_residues(DummyGroup([2, 1])),
                           np.array([self.values[[2, 3, 9]].sum(),
                                     self.values[[4, 5, 8]].sum()]))

    def test_get_segments(self, attr):
        assert_equal(attr.get_segments(DummyGroup([1])),
                           np.array([self.values[[4, 5, 8, 2, 3, 9]].sum()]))

    def test_get_segment(self, attr):
        assert_equal(attr.get_segments(DummyGroup(1)),
                           np.sum(self.values[[4, 5, 8, 2, 3, 9]]))


class TestMasses(AggregationMixin):
    attrclass = tpattrs.Masses


class TestCharges(AggregationMixin):
    values = np.array([+2, -1, 0, -1, +1, +2, 0, 0, 0, -1])
    attrclass = tpattrs.Charges


class TestResidueAttr(TopologyAttrMixin):
    """Test residue-level TopologyAttrs.

    """
    single_value = 2
    values = np.array([15.2, 395.6, 0.1, 9.8])
    attrclass = tpattrs.ResidueAttr

    def test_set_residue_VE(self, universe):
        # setting e.g. resname to 2 values should fail with VE
        res = universe.residues[0]
        with pytest.raises(ValueError):
            setattr(res, self.attrclass.singular, self.values[:2])

    def test_get_atoms(self, attr):
        assert_equal(attr.get_atoms(DummyGroup([7, 3, 9])),
                     self.values[[3, 2, 2]])

    def test_get_atom(self, universe):
        attr = getattr(universe.atoms[0], self.attrclass.singular)
        assert_equal(attr, self.values[0])

    def test_get_residues(self, attr):
        assert_equal(attr.get_residues(DummyGroup([1, 2, 1, 3])),
                     self.values[[1, 2, 1, 3]])

    def test_set_residues_singular(self, attr):
        dg = DummyGroup([3, 0, 1])
        attr.set_residues(dg, self.single_value)

        assert_equal(attr.get_residues(dg),
                     np.array([self.single_value]*3, dtype=self.values.dtype))

    def test_set_residues_plural(self, attr):
        attr.set_residues(DummyGroup([3, 0, 1]),
                               np.array([23, 504, 2]))
        assert_almost_equal(attr.get_residues(DummyGroup([3, 0, 1])),
                                  np.array([23, 504, 2]))

    def test_set_residues_VE(self, attr):
        dg = DummyGroup([3, 0, 1])

        with pytest.raises(ValueError):
            attr.set_residues(dg, np.array([4.5, 5.2]))

    def test_get_segments(self, attr):
        """Unless overriden by child class, this should yield values for all
        atoms in segments.

        """
        assert_equal(attr.get_segments(DummyGroup([0, 1, 1])),
                           [self.values[[0, 3]], self.values[[1, 2]], self.values[[1, 2]]])


class TestResnames(TestResidueAttr):
    attrclass = tpattrs.Resnames
    single_value = 'xyz'
    values = np.array(['a', 'b', '', 'd'], dtype=object)


class TestICodes(TestResnames):
    attrclass = tpattrs.ICodes


class TestResids(TestResidueAttr):
    values = np.array([10, 11, 18, 20])
    attrclass = tpattrs.Resids

    @pytest.mark.xfail
    def test_set_atoms(self, attr):
        """Setting the resids of atoms changes their residue membership.

        """
        # moving resids doesn't currently work!
        assert 1 == 2

        # set with array
        attr.set_atoms(DummyGroup([3, 7]), np.array([11, 20]))
        assert_equal(attr.get_atoms(DummyGroup([3, 7])), np.array([11, 20]))

        # set to resid that no residue has (should raise exception)
        with pytest.raises(NoDataError):
            attr.set_atoms(DummyGroup([3, 7]), np.array([11, 21]))

    def test_set_residues(self, attr):
        attr.set_residues(DummyGroup([3, 0, 1]),
                               np.array([23, 504, 27]))
        assert_almost_equal(attr.get_residues(DummyGroup([3, 0, 1])),
                                  np.array([23, 504, 27]))


class TestSegmentAttr(TopologyAttrMixin):
    """Test segment-level TopologyAttrs.

    """
    values = np.array([-0.19, 500])
    attrclass = tpattrs.SegmentAttr

    def test_set_segment_VE(self):
        u = make_Universe(('segids',))
        seg = u.segments[0]
        with pytest.raises(ValueError):
            setattr(seg, 'segid', [1, 2, 3])

    def test_get_atoms(self, attr):
        assert_equal(attr.get_atoms(DummyGroup([2, 4, 1])),
                           self.values[[1, 1, 0]])

    def test_get_residues(self, attr):
        assert_equal(attr.get_residues(DummyGroup([1, 2, 1, 3])),
                           self.values[[1, 1, 1, 0]])

    def test_get_segments(self, attr):
        """Unless overriden by child class, this should yield values for all
        atoms in segments.

        """
        assert_equal(attr.get_segments(DummyGroup([1, 0, 0])),
                           self.values[[1, 0, 0]])

    def test_set_segments_singular(self, attr):
        dg = DummyGroup([0, 1])
        attr.set_segments(dg, 0.45)
        assert_equal(attr.get_segments(dg), np.array([0.45, 0.45]))

    def test_set_segments_plural(self, attr):
        dg = DummyGroup([0, 1])
        attr.set_segments(dg, np.array([23, -0.0002]))
        assert_equal(attr.get_segments(dg), np.array([23, -0.0002]))

    def test_set_segments_VE(self, attr):
        dg = DummyGroup([0, 1])
        with pytest.raises(ValueError):
            attr.set_segments(dg, np.array([4, 5, 6, 7]))


class TestAttr(object):
    @pytest.fixture()
    def ag(self):
        universe = mda.Universe(PSF, DCD)
        return universe.atoms  # prototypical AtomGroup

    def test_principal_axes(self, ag):
        assert_almost_equal(
            ag.principal_axes(),
            np.array([
                      [1.53389276e-03, 4.41386224e-02, 9.99024239e-01],
                      [1.20986911e-02, 9.98951474e-01, -4.41539838e-02],
                      [-9.99925632e-01, 1.21546132e-02, 9.98264877e-04]]))

    @pytest.fixture()
    def universe_pa(self):
        return mda.Universe(PDB_CHECK_RIGHTHAND_PA)

    def test_principal_axes_handedness(self, universe_pa):
        e_vec = universe_pa.atoms.principal_axes()
        assert_almost_equal(np.dot(np.cross(e_vec[0], e_vec[1]), e_vec[2]), 1.0)

    def test_align_principal_axes_with_self(self, ag):
        pa = ag.principal_axes()
        ag.align_principal_axis(0, pa[0])

        assert_almost_equal(ag.principal_axes(), pa)

    def test_align_principal_axes_with_x(self, ag):
        ag.align_principal_axis(0, [1, 0, 0])
        # This is a very loose check that the difference is not more then 0.5.
        # This is OK here because the rounding error in the calculation really
        # is that big.
        assert_almost_equal(np.abs(ag.principal_axes()), np.eye(3), decimal=1)


class TestCrossLevelAttributeSetting(object):
    """

    Can only get attributes belonging to higher level objects

    Atom.resid works!
    ResidueGroup.names = ['a', 'b', 'c'] doesn't work, Atom is below Residue

    Setting any attribute we can get should only work if they are the same level.

    Atom.resid = 4  should fail because resid belongs to Residue not Atom
    """

    u = make_Universe(('names', 'resids', 'segids'))

    # component and group in each level
    atomlevel = (u.atoms[0], u.atoms[:10])
    residuelevel = (u.residues[0], u.residues[:5])
    segmentlevel = (u.segments[0], u.segments[:2])
    levels = {0: atomlevel, 1: residuelevel, 2: segmentlevel}

    atomattr = 'names'
    residueattr = 'resids'
    segmentattr = 'segids'
    attrs = {0: atomattr, 1: residueattr, 2: segmentattr}

    @pytest.mark.parametrize('level_idx, level', levels.items())
    @pytest.mark.parametrize('attr_idx, attr', attrs.items())
    def test_set_crosslevel(self, level_idx, level, attr_idx, attr):
        if level_idx == attr_idx:
            # if we're on the same level, then this should work
            # ie Atom.mass = 12.0 is OK!
            return
        component, group = level
        # eg 'name', 'names' = 'names', 'names'[:-1]
        singular_attr, plural_attr = attr[:-1], attr

        # eg check ResidueGroup.names = 'newvalue' raises NIE
        # or ResidueGroup.segids = 'newvalue' raises NIE
        self._check_crosslevel_fail(group, plural_attr)

        if attr_idx < level_idx:
            # Segment.resid doesn't even exist as an attribute
            # so we don't have to check that setting fails
            # Atom.segid does exist as attribute,
            # but will fail to be set
            return
        self._check_crosslevel_fail(component, singular_attr)

    @staticmethod
    def _check_crosslevel_fail(item, attr):
        with pytest.raises(NotImplementedError):
            setattr(item, attr, 1.0)


class TestRecordTypes(object):
    def test_record_types_default(self):
        u = make_Universe()

        u.add_TopologyAttr('record_type')

        assert u.atoms[0].record_type == 'ATOM'
        assert_equal(u.atoms[:10].record_types, 'ATOM')

    @pytest.fixture()
    def rectype_uni(self):
        # standard 125/25/5 universe
        u = make_Universe()
        u.add_TopologyAttr('record_type')
        # first 25 atoms are ATOM (first 5 residues, first segment)
        # 25 to 50th are HETATM (res 5:10, second segment)
        # all after are ATOM
        u.atoms[:25].record_types = 'ATOM'
        u.atoms[25:50].record_types = 'HETATM'
        u.atoms[50:].record_types = 'ATOM'

        return u

    def test_encoding(self, rectype_uni):
        ag = rectype_uni.atoms[:10]

        ag[0].record_type = 'ATOM'
        ag[1:4].record_types = 'HETATM'

        assert ag[0].record_type == 'ATOM'
        assert ag[1].record_type == 'HETATM'

    def test_residue_record_types(self, rectype_uni):
        rt = rectype_uni.residues.record_types

        assert isinstance(rt, list)
        assert len(rt) == 25

        # check return type explicitly
        # some versions of numpy allow bool to str comparison
        assert not rt[0].dtype == bool
        assert (rt[0] == 'ATOM').all()
        assert (rt[5] == 'HETATM').all()

    def test_segment_record_types(self, rectype_uni):
        rt = rectype_uni.segments.record_types

        assert isinstance(rt, list)
        assert len(rt) == 5

        assert not rt[0].dtype == bool
        assert (rt[0] == 'ATOM').all()
        assert (rt[1] == 'HETATM').all()


def test_static_typing():
    ta = tpattrs.Charges(['1.0', '2.0', '3.0'])

    assert isinstance(ta.values, np.ndarray)
    assert ta.values.dtype == float


def test_static_typing_from_empty():
    u = mda.Universe.empty(3)

    u.add_TopologyAttr('masses', values=['1.0', '2.0', '3.0'])

    assert isinstance(u._topology.masses.values, np.ndarray)
    assert isinstance(u.atoms[0].mass, float)


@pytest.mark.parametrize('level, transplant_name', (
    ('atoms', 'center_of_mass'),
    ('atoms', 'center_of_charge'),
    ('atoms', 'total_charge'),
    ('residues', 'total_charge'),
))
def test_stub_transplant_methods(level, transplant_name):
    u = mda.Universe.empty(n_atoms=2)
    group = getattr(u, level)
    with pytest.raises(NoDataError):
        getattr(group, transplant_name)()


@pytest.mark.parametrize('level, transplant_name', (
    ('universe', 'models'),
    ('atoms', 'n_fragments'),
))
def test_stub_transplant_property(level, transplant_name):
    u = mda.Universe.empty(n_atoms=2)
    group = getattr(u, level)
    with pytest.raises(NoDataError):
        getattr(group, transplant_name)


def test_warn_selection_for_strange_dtype():
    err = "A selection keyword could not be automatically generated"

    with pytest.warns(UserWarning, match=err):
        class Star(tpattrs.TopologyAttr):
            singular = "star"  # turns out test_imports doesn't like emoji
            attrname = "stars"  # :(
            per_object = "atom"
            dtype = dict


class TestDeprecateBFactor:

    MATCH = "use the tempfactor attribute"

    @pytest.fixture()
    def universe(self):
        return mda.Universe(MMTF)

    def test_deprecate_bfactors_get_group(self, universe):
        with pytest.warns(DeprecationWarning, match=self.MATCH):
            universe.atoms.bfactors

    def test_deprecate_bfactors_get_atom(self, universe):
        with pytest.warns(DeprecationWarning, match=self.MATCH):
            assert universe.atoms[0].bfactor == universe.atoms[0].tempfactor

    def test_deprecate_bfactors_set_group(self, universe):
        with pytest.warns(DeprecationWarning, match=self.MATCH):
            universe.atoms[:2].bfactors = [3.14, 10]
        assert universe.atoms.tempfactors[0] == 3.14
        assert universe.atoms.tempfactors[1] == 10

        with pytest.warns(DeprecationWarning, match=self.MATCH):
            assert universe.atoms.bfactors[0] == 3.14
            assert universe.atoms.bfactors[1] == 10

    def test_deprecate_bfactors_set_atom(self, universe):
        with pytest.warns(DeprecationWarning, match=self.MATCH):
            universe.atoms[0].bfactor = 3.14
        assert universe.atoms[0].tempfactor == 3.14
        with pytest.warns(DeprecationWarning, match=self.MATCH):
            assert universe.atoms[0].bfactor == 3.14

    def test_deprecate_bfactor_sel(self, universe):
        with pytest.warns(DeprecationWarning, match=self.MATCH):
            universe.select_atoms("bfactor 3")


class TestStringInterning:
    # try and trip up the string interning we use for string attributes
    @pytest.fixture
    def universe(self):
        u = mda.Universe.empty(n_atoms=10, n_residues=2,
                               atom_resindex=[0]*5 + [1] * 5)
        u.add_TopologyAttr('names', values=['A'] * 10)
        u.add_TopologyAttr('resnames', values=['ResA', 'ResB'])
        u.add_TopologyAttr('segids', values=['SegA'])

        return u

    @pytest.mark.parametrize('newname', ['ResA', 'ResB'])
    def test_add_residue(self, universe, newname):
        newres = universe.add_Residue(resname=newname)

        assert newres.resname == newname

        ag = universe.atoms[2]
        ag.residue = newres

        assert ag.resname == newname

    @pytest.mark.parametrize('newname', ['SegA', 'SegB'])
    def test_add_segment(self, universe, newname):
        newseg = universe.add_Segment(segid=newname)

        assert newseg.segid == newname

        rg = universe.residues[0]
        rg.segment = newseg

        assert rg.atoms[0].segid == newname

    def test_issue3437(self, universe):
        newseg = universe.add_Segment(segid='B')

        ag = universe.residues[0].atoms

        ag.residues.segments = newseg

        assert 'B' in universe.segments.segids

        ag2 = universe.select_atoms('segid B')

        assert len(ag2) == 5
        assert (ag2.ix == ag.ix).all()


class Testcenter_of_charge():

    compounds = ['group', 'segments', 'residues', 'molecules', 'fragments']

    @pytest.fixture
    def u(self):
        """A universe containing two dimers with a finite dipole moment."""
        universe = mda.Universe.empty(n_atoms=4,
                                      n_residues=2,
                                      n_segments=2,
                                      atom_resindex=[0, 0, 1, 1],
                                      residue_segindex=[0, 1])

        universe.add_TopologyAttr("masses", [1, 0, 0, 1])
        universe.add_TopologyAttr("charges", [1, -1, -1, 1])
        universe.add_TopologyAttr("bonds", ((0, 1), (2, 3)))
        universe.add_TopologyAttr("resids", [0, 1])
        universe.add_TopologyAttr("molnums", [0, 1])

        positions = np.array([[0, 0, 0], [0, 1, 0], [2, 1, 0], [2, 2, 0]])

        universe.trajectory = get_reader_for(positions)(positions,
                                                        order='fac',
                                                        n_atoms=4)

        for ts in universe.trajectory:
            ts.dimensions = np.array([1, 2, 3, 90, 90, 90])

        return universe

    @pytest.mark.parametrize('compound', compounds)
    def test_coc(self, u, compound):
        coc = u.atoms.center_of_charge(compound=compound)
        if compound == "group":
            coc_ref = [1, 1, 0]
        else:
            coc_ref = [[0, 0.5, 0], [2, 1.5, 0]]
        assert_equal(coc, coc_ref)

    @pytest.mark.parametrize('compound', compounds)
    def test_coc_wrap(self, u, compound):
        coc = u.atoms[:2].center_of_charge(compound=compound, wrap=True)
        assert_equal(coc.flatten(), [0, 0.5, 0])

    @pytest.mark.parametrize('compound', compounds)
    def test_coc_unwrap(self, u, compound):
        u.atoms.wrap
        coc = u.atoms[:2].center_of_charge(compound=compound, unwrap=True)
        assert_equal(coc.flatten(), [0, -0.5, 0])



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import numpy as np
import pytest

from numpy.testing import assert_equal

from MDAnalysis.lib.mdamath import triclinic_vectors

from MDAnalysisTests.datafiles import (DMS)


class TestDMSReader(object):
    @pytest.fixture()
    def universe(self):
        return mda.Universe(DMS)

    @pytest.fixture()
    def ts(self, universe):
        return universe.trajectory.ts

    def test_global_cell(self, ts):
        assert ts.dimensions is None

    def test_velocities(self, ts):
        assert_equal(hasattr(ts, "_velocities"), False)

    def test_number_of_coords(self, universe):
        # Desired value taken from VMD
        #      Info)    Atoms: 3341
        assert_equal(len(universe.atoms), 3341)

    def test_coords_atom_0(self, universe):
        # Desired coordinates taken directly from the SQLite file. Check unit
        # conversion
        coords_0 = np.array([-11.0530004501343,
                             26.6800003051758,
                             12.7419996261597, ],
                            dtype=np.float32)
        assert_equal(universe.atoms[0].position, coords_0)

    def test_n_frames(self, universe):
        assert_equal(universe.trajectory.n_frames, 1,
                     "wrong number of frames in pdb")

    def test_time(self, universe):
        assert_equal(universe.trajectory.time, 0.0,
                     "wrong time of the frame")

    def test_frame(self, universe):
        assert_equal(universe.trajectory.frame, 0, "wrong frame number "
                                                   "(0-based, should be 0 for single frame readers)")

    def test_frame_index_0(self, universe):
        universe.trajectory[0]
        assert_equal(universe.trajectory.ts.frame, 0,
                     "frame number for frame index 0 should be 0")

    def test_frame_index_1_raises_IndexError(self, universe):
        with pytest.raises(IndexError):
            universe.trajectory[1]



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2018 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

from distutils.util import strtobool
import os

import pytest

from collections import defaultdict, Counter
from numpy.testing import assert_equal, assert_allclose
import numpy as np

import MDAnalysis as mda
from MDAnalysisTests.datafiles import (
    GRO, Martini_membrane_gro, PDB, PDB_xvf, SURFACE_PDB, SURFACE_TRR
)
from MDAnalysis.lib import nsgrid
from MDAnalysis.transformations.translate import center_in_box


@pytest.fixture
def universe():
    u = mda.Universe(GRO)
    return u

def run_grid_search(u, ref_id, cutoff=3):
    coords = u.atoms.positions
    searchcoords = u.atoms.positions[ref_id]
    if searchcoords.shape == (3, ):
        searchcoords = searchcoords[None, :]
    # Run grid search
    searcher = nsgrid.FastNS(cutoff, coords, box=u.dimensions)

    return searcher.search(searchcoords)

@pytest.mark.parametrize('box', [
    np.zeros(3),  # Bad shape
    np.zeros((3, 3)),  # Collapsed box
    np.array([[0, 0, 0], [0, 1, 0], [0, 0, 1]]),  # 2D box
    np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),  # Box provided as array of integers
    np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float64),  # Box provided as array of double
])
def test_pbc_box(box):
    """Check that PBC box accepts only well-formated boxes"""
    coords = np.array([[1.0, 1.0, 1.0]], dtype=np.float32)

    with pytest.raises(ValueError):
        nsgrid.FastNS(4.0, coords, box=box)


@pytest.mark.parametrize('cutoff, match', ((-4, "Cutoff must be positive"),
                                           (100000,
                                            "Cutoff 100000 too large for box")))
def test_nsgrid_badcutoff(universe, cutoff, match):
    with pytest.raises(ValueError, match=match):
        run_grid_search(universe, 0, cutoff)


def test_ns_grid_noneighbor(universe):
    """Check that grid search returns empty lists/arrays when there is no neighbors"""
    ref_id = 0
    cutoff = 0.5

    results_grid = run_grid_search(universe, ref_id, cutoff)

    # same indices will be selected as neighbour here
    assert len(results_grid.get_pairs()) == 1
    assert len(results_grid.get_pair_distances()) == 1


def test_nsgrid_PBC_rect():
    """Check that nsgrid works with rect boxes and PBC"""
    ref_id = 191
    # Atomid are from gmx select so there start from 1 and not 0. hence -1!
    results = np.array([191, 192, 672, 682, 683, 684, 995, 996, 2060, 2808, 3300, 3791,
                        3792]) - 1

    universe = mda.Universe(Martini_membrane_gro)
    cutoff = 7

    # FastNS is called differently to max coverage
    searcher = nsgrid.FastNS(cutoff, universe.atoms.positions, box=universe.dimensions)

    results_grid = searcher.search(universe.atoms.positions[ref_id][None, :]).get_pairs()
    other_ix = sorted(i for (_, i) in results_grid)

    assert len(results) == len(results_grid)
    assert other_ix == sorted(results)


def test_nsgrid_PBC(universe):
    """Check that grid search works when PBC is needed"""
    # Atomid are from gmx select so there start from 1 and not 0. hence -1!
    ref_id = 13937
    results = np.array([4398, 4401, 13938, 13939, 13940, 13941, 17987, 23518, 23519, 23521, 23734,
                        47451]) - 1

    results_grid = run_grid_search(universe, ref_id).get_pairs()

    other_ix = sorted(i for (_, i) in results_grid)

    assert len(results) == len(other_ix)
    assert other_ix == sorted(results)


def test_nsgrid_pairs(universe):
    """Check that grid search returns the proper pairs"""

    ref_id = 13937
    neighbors = np.array([4398, 4401, 13938, 13939, 13940, 13941, 17987, 23518, 23519, 23521, 23734,
                          47451]) - 1  # Atomid are from gmx select so there start from 1 and not 0. hence -1!
    results = []

    results = np.array(results)

    results_grid = run_grid_search(universe, ref_id).get_pairs()

    assert_equal(np.sort(neighbors, axis=0), np.sort(results_grid[:, 1], axis=0))


def test_nsgrid_pair_distances(universe):
    """Check that grid search returns the proper pair distances"""

    ref_id = 13937
    results = np.array([0.0, 0.270, 0.285, 0.096, 0.096, 0.015, 0.278, 0.268, 0.179, 0.259, 0.290,
                        0.270]) * 10  # These distances where obtained by gmx distance so they are in nm

    results_grid = run_grid_search(universe, ref_id).get_pair_distances()

    assert_allclose(np.sort(results), np.sort(results_grid), atol=1e-2)


def test_nsgrid_distances(universe):
    """Check that grid search returns the proper distances"""
    # These distances where obtained by gmx distance so they are in nm
    ref_id = 13937
    results = np.array([0.0, 0.270, 0.285, 0.096, 0.096, 0.015, 0.278, 0.268, 0.179, 0.259, 0.290,
                        0.270]) * 10

    results_grid = run_grid_search(universe, ref_id).get_pair_distances()

    assert_allclose(np.sort(results), np.sort(results_grid), atol=1e-2)


@pytest.mark.parametrize('box, results',
                         ((None, [3, 13, 24]),
                          (np.array([10., 10., 10., 90., 90., 90.]), [3, 13, 24, 39, 67]),
                          (np.array([10., 10., 10., 60., 75., 90.]), [3, 13, 24, 39, 60, 79])))
def test_nsgrid_search(box, results):
    np.random.seed(90003)
    points = (np.random.uniform(low=0, high=1.0,
                        size=(100, 3))*(10.)).astype(np.float32)
    cutoff = 2.0
    query = np.array([1., 1., 1.], dtype=np.float32).reshape((1, 3))

    if box is None:
        pseudobox = np.zeros(6, dtype=np.float32)
        all_coords = np.concatenate([points, query])
        lmax = all_coords.max(axis=0)
        lmin = all_coords.min(axis=0)
        pseudobox[:3] = 1.1*(lmax - lmin)
        pseudobox[3:] = 90.
        shiftpoints, shiftquery = points.copy(), query.copy()
        shiftpoints -= lmin
        shiftquery -= lmin
        searcher = nsgrid.FastNS(cutoff, shiftpoints, box=pseudobox, pbc=False)
        searchresults = searcher.search(shiftquery)
    else:
        searcher = nsgrid.FastNS(cutoff, points, box)
        searchresults = searcher.search(query)
    indices = searchresults.get_pairs()[:, 1]
    assert_equal(np.sort(indices), results)


@pytest.mark.parametrize('box, result',
                         ((None, 21),
                          (np.array([0., 0., 0., 90., 90., 90.]), 21),
                          (np.array([10., 10., 10., 90., 90., 90.]), 26),
                          (np.array([10., 10., 10., 60., 75., 90.]), 33)))
def test_nsgrid_selfsearch(box, result):
    np.random.seed(90003)
    points = (np.random.uniform(low=0, high=1.0,
                        size=(100, 3))*(10.)).astype(np.float32)
    cutoff = 1.0
    if box is None or np.allclose(box[:3], 0):
        # create a pseudobox
        # define the max range
        # and supply the pseudobox
        # along with only one set of coordinates
        pseudobox = np.zeros(6, dtype=np.float32)
        lmax = points.max(axis=0)
        lmin = points.min(axis=0)
        pseudobox[:3] = 1.1*(lmax - lmin)
        pseudobox[3:] = 90.
        shiftref = points.copy()
        shiftref -= lmin
        searcher = nsgrid.FastNS(cutoff, shiftref, box=pseudobox, pbc=False)
        searchresults = searcher.self_search()
    else:
        searcher = nsgrid.FastNS(cutoff, points, box=box)
        searchresults = searcher.self_search()
    pairs = searchresults.get_pairs()
    assert_equal(len(pairs), result)

def test_nsgrid_probe_close_to_box_boundary():
    # FastNS.search used to segfault with this box, cutoff and reference
    # coordinate prior to PR #2136, so we ensure that this remains fixed.
    # See Issue #2132 for further information.
    ref = np.array([[55.783722, 44.190044, -54.16671]], dtype=np.float32)
    box = np.array([53.785854, 43.951054, 57.17597, 90., 90., 90.], dtype=np.float32)
    cutoff = 3.0
    # search within a configuration where we know the expected outcome:
    conf = np.ones((1, 3), dtype=np.float32)
    searcher = nsgrid.FastNS(cutoff, conf, box)
    results = searcher.search(ref)
    # check if results are as expected:
    expected_pairs = np.zeros((1, 2), dtype=np.int64)
    expected_dists = np.array([2.3689647], dtype=np.float64)
    assert_equal(results.get_pairs(), expected_pairs)
    assert_allclose(results.get_pair_distances(), expected_dists, rtol=1.e-6)


def test_zero_max_dist():
    # see issue #2656
    # searching with max_dist = 0.0 shouldn't cause segfault (and infinite subboxes)
    ref = np.array([1.0, 1.0, 1.0], dtype=np.float32)
    conf = np.array([2.0, 1.0, 1.0], dtype=np.float32)

    box = np.array([10., 10., 10., 90., 90., 90.], dtype=np.float32)

    res = mda.lib.distances._nsgrid_capped(ref, conf, box=box, max_cutoff=0.0)


@pytest.fixture()
def u_pbc_triclinic():
    u = mda.Universe(PDB)
    u.dimensions = [10, 10, 10, 60, 60, 60]
    return u


def test_around_res(u_pbc_triclinic):
    # sanity check for issue 2656, shouldn't segfault (obviously)
    ag = u_pbc_triclinic.select_atoms('around 0.0 resid 3')
    assert len(ag) == 0


def test_around_overlapping():
    # check that around 0.0 catches when atoms *are* superimposed
    u = mda.Universe.empty(60, trajectory=True)
    xyz = np.zeros((60, 3))
    x = np.tile(np.arange(12), (5,))+np.repeat(np.arange(5)*100, 12)
    # x is 5 images of 12 atoms

    xyz[:, 0] = x  # y and z are 0
    u.load_new(xyz)

    u.dimensions = [100, 100, 100, 60, 60, 60]
    # Technically true but not what we're testing:
    # dist = mda.lib.distances.distance_array(u.atoms[:12].positions,
    #                                         u.atoms[12:].positions,
    #                                         box=u.dimensions)
    # assert np.count_nonzero(np.any(dist <= 0.0, axis=0)) == 48
    assert u.select_atoms('around 0.0 index 0:11').n_atoms == 48


def test_issue_2229_part1():
    # reproducing first case in GH issue 2229
    u = mda.Universe.empty(2, trajectory=True)

    u.dimensions = [57.45585, 50.0000, 50.0000, 90, 90, 90]

    u.atoms[0].position = [0, 0, 0]
    u.atoms[1].position = [55.00, 0, 0]

    g = mda.lib.nsgrid.FastNS(3.0, u.atoms[[0]].positions, box=u.dimensions)
    assert len(g.search(u.atoms[[1]].positions).get_pairs()) == 1

    g = mda.lib.nsgrid.FastNS(3.0, u.atoms[[1]].positions, box=u.dimensions)
    assert len(g.search(u.atoms[[0]].positions).get_pairs()) == 1


def test_issue_2229_part2():
    u = mda.Universe.empty(2, trajectory=True)

    u.dimensions = [45.0000, 55.0000, 109.8375, 90, 90, 90]

    u.atoms[0].position = [0, 0, 29.29]
    u.atoms[1].position = [0, 0, 28.23]

    g = mda.lib.nsgrid.FastNS(3.0, u.atoms[[0]].positions, box=u.dimensions, pbc=False)
    assert len(g.search(u.atoms[[1]].positions).get_pairs()) == 1

    g = mda.lib.nsgrid.FastNS(3.0, u.atoms[[1]].positions, box=u.dimensions)
    assert len(g.search(u.atoms[[0]].positions).get_pairs()) == 1


def test_issue_2919():
    # regression test reported in issue 2919
    # other methods will also give 1115 or 2479 results
    u = mda.Universe(PDB_xvf)
    ag = u.select_atoms('index 0')
    u.trajectory.ts = center_in_box(ag)(u.trajectory.ts)

    box = u.dimensions
    reference = u.select_atoms('protein')
    configuration = u.select_atoms('not protein')

    for cutoff, expected in [(2.8, 1115), (3.2, 2497)]:
        pairs, distances = mda.lib.distances.capped_distance(
            reference.positions,
            configuration.positions,
            max_cutoff=cutoff,
            box=box,
            method='nsgrid',
            return_distances=True,
        )
        assert len(pairs) == expected


def test_issue_2345():
    # another example of NSGrid being wrong
    # this is a 111 FCC slab
    # coordination numbers for atoms should be either 9 or 12, 50 of each
    u = mda.Universe(SURFACE_PDB, SURFACE_TRR)

    g = mda.lib.nsgrid.FastNS(2.9, u.atoms.positions, box=u.dimensions)

    cn = defaultdict(list)

    idx = g.self_search().get_pairs()
    # count number of contacts for each atom
    for (i, j) in idx:
        cn[i].append(j)
        cn[j].append(i)
    c = Counter(len(v) for v in cn.values())

    assert c == {9: 50, 12: 50}


def test_issue_2670():
    # Tests that NSGrid no longer crashes when using small box sizes
    u = mda.Universe(PDB)
    u.dimensions = [1e-3, 1e-3, 1e-3, 90, 90, 90]

    # PDB files only have a coordinate precision of 1.0e-3, so we need to scale
    # the coordinates for this test to make any sense:
    u.atoms.positions = u.atoms.positions * 1.0e-3

    ag1 = u.select_atoms('resid 2 3')
    # should return nothing as nothing except resid 3 is within 0.0 or resid 3
    assert len(ag1.select_atoms('around 0.0 resid 3')) == 0

    # force atom 0 of resid 1 to overlap with atom 0 of resid 3
    u.residues[0].atoms[0].position = u.residues[2].atoms[0].position
    ag2 = u.select_atoms('resid 1 3')

    # should return the one atom overlap
    assert len(ag2.select_atoms('around 0.0 resid 3')) == 1


def high_mem_tests_enabled():
    """ Returns true if ENABLE_HIGH_MEM_UNIT_TESTS is set to true."""
    env = os.getenv("ENABLE_HIGH_MEM_UNIT_TESTS", default="0")
    try:
        return strtobool(env)
    except ValueError:
        return False


reason = ("Turned off by default. The test can be enabled by setting "
          "the ENABLE_HIGH_MEM_UNIT_TESTS "
          "environment variable. Make sure you have at least 10GB of RAM.")


# Tests that with a tiny cutoff to box ratio, the number of grids is capped
# to avoid indexing overflow. Expected results copied from test_nsgrid_search
# with no box.
@pytest.mark.skipif(not high_mem_tests_enabled(), reason=reason)
def test_issue_3183():
    np.random.seed(90003)
    points = (np.random.uniform(low=0, high=1.0,
                                size=(100, 3)) * (10.)).astype(np.float32)
    cutoff = 2.0
    query = np.array([1., 1., 1.], dtype=np.float32).reshape((1, 3))
    box = np.array([10000., 10000., 10000., 90., 90., 90.])

    searcher = nsgrid.FastNS(cutoff, points, box)
    searchresults = searcher.search(query)
    indices = searchresults.get_pairs()[:, 1]
    want_results = [3, 13, 24]
    assert_equal(np.sort(indices), want_results)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from numpy.testing import assert_equal
import pytest

import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    DLP_CONFIG,
    DLP_CONFIG_order,
    DLP_CONFIG_minimal,
    DLP_HISTORY,
    DLP_HISTORY_order,
    DLP_HISTORY_minimal,
    DLP_HISTORY_minimal_cell,
    DLP_HISTORY_classic
)


class DLPUniverse(ParserBase):
    def test_creates_universe(self, filename):
        u = mda.Universe(filename, topology_format=self.format)
        assert isinstance(u, mda.Universe)


class DLPBase2(DLPUniverse):
    expected_attrs = ['ids', 'names']
    guessed_attrs = ['types', 'masses']
    expected_n_atoms = 216
    expected_n_residues = 1
    expected_n_segments = 1

    def test_names(self, top):
        assert top.names.values[0] == 'K+'
        assert top.names.values[4] == 'Cl-'


class TestDLPHistoryParser(DLPBase2):
    parser = mda.topology.DLPolyParser.HistoryParser
    ref_filename = DLP_HISTORY
    format = 'HISTORY'


class TestDLPConfigParser(DLPBase2):
    parser = mda.topology.DLPolyParser.ConfigParser
    ref_filename = DLP_CONFIG
    format = 'CONFIG'


class DLPBase(DLPUniverse):
    expected_attrs = ['ids', 'names']
    guessed_attrs = ['types', 'masses']
    expected_n_atoms = 3
    expected_n_residues = 1
    expected_n_segments = 1

    def test_dlp_names(self, top):
        assert_equal(top.names.values,
                     ['C', 'B', 'A'])


class TestDLPConfigMinimal(DLPBase):
    parser = mda.topology.DLPolyParser.ConfigParser
    ref_filename = DLP_CONFIG_minimal
    format = 'CONFIG'


class TestDLPConfigOrder(DLPBase):
    parser = mda.topology.DLPolyParser.ConfigParser
    ref_filename = DLP_CONFIG_order
    format = 'CONFIG'


class TestDLPHistoryMinimal(DLPBase):
    parser = mda.topology.DLPolyParser.HistoryParser
    ref_filename = DLP_HISTORY_minimal
    format = 'HISTORY'


class TestDLPHistoryMinimal(DLPBase):
    parser = mda.topology.DLPolyParser.HistoryParser
    ref_filename = DLP_HISTORY_minimal_cell
    format = 'HISTORY'


class TestDLPHistoryOrder(DLPBase):
    parser = mda.topology.DLPolyParser.HistoryParser
    ref_filename = DLP_HISTORY_order
    format = 'HISTORY'


class TestDLPHistoryClassic(DLPBase):
    parser = mda.topology.DLPolyParser.HistoryParser
    ref_filename = DLP_HISTORY_classic
    format = 'HISTORY'


def test_HISTORY_EOFError():
    with pytest.raises(EOFError):
        mda.Universe(DLP_CONFIG, topology_format='HISTORY')



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from contextlib import contextmanager

import MDAnalysis as mda
import MDAnalysis.analysis.align as align
import MDAnalysis.analysis.rms as rms
import os
import numpy as np
import pytest
from MDAnalysis import SelectionError, SelectionWarning
from MDAnalysisTests import executable_not_found
from MDAnalysisTests.datafiles import (PSF, DCD, CRD, FASTA, ALIGN_BOUND,
                                       ALIGN_UNBOUND, PDB_helix)
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
    assert_array_equal,
    assert_array_almost_equal,
    assert_allclose,
)

#Function for Parametrizing conditional raising
@contextmanager
def does_not_raise():
    yield

class TestRotationMatrix(object):
    a = np.array([[0.1, 0.2, 0.3], [1.1, 1.1, 1.1]])
    b = np.array([[0.1, 0.1, 0.1], [1.1, 1.1, 1.1]])
    w = np.array([1.3, 2.3])

    @pytest.mark.parametrize('a, b, weights', (
            (a, b, None),
            (a, b, w),
            (a.astype(int), b.astype(int), w.astype(np.float32))
    ))
    def test_rotation_matrix_input(self, a, b, weights):
        rot, rmsd = align.rotation_matrix(a, b, weights)
        assert_equal(rot, np.eye(3))
        assert rmsd is None

    def test_list_args(self):
        a = [[0.1, 0.2, 0.3], [1.1, 1.1, 1.1]]
        b = [[0.1, 0.1, 0.1], [1.1, 1.1, 1.1]]
        w = [1.3, 2.3]
        rot, rmsd = align.rotation_matrix(a, b, w)
        assert_equal(rot, np.eye(3))
        assert rmsd is None

    def test_exception(self):
        a = [[0.1, 0.2, 0.3],
             [1.1, 1.1, 1.1],
             [2, 2, 2]]
        b = [[0.1, 0.1, 0.1],
             [1.1, 1.1, 1.1]]
        with pytest.raises(ValueError):
            align.rotation_matrix(a, b)


class TestGetMatchingAtoms(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PSF, DCD)

    @staticmethod
    @pytest.fixture()
    def reference():
        return mda.Universe(PSF, DCD)

    @staticmethod
    @pytest.fixture()
    def reference_small(reference):
        return mda.Merge(reference.select_atoms(
            "not name H* and not atom 4AKE 1 CA"))

    @pytest.mark.parametrize("strict", (True, False))
    def test_match(self, universe, reference, strict,
                   selection="protein and backbone"):
        ref = reference.select_atoms(selection)
        mobile = universe.select_atoms(selection)
        groups = align.get_matching_atoms(ref, mobile, strict=strict)
        assert_equal(groups[0].names, groups[1].names)

    @pytest.mark.parametrize("strict", (True, False))
    def test_nomatch_atoms_raise(self, universe, reference,
                                 strict, selection="protein and backbone"):
        # one atom less but same residues; with strict=False should try
        # to get selections (but current code fails, so we also raise SelectionError)
        ref = reference.select_atoms(selection).atoms[1:]
        mobile = universe.select_atoms(selection)
        if strict:
            with pytest.raises(SelectionError):
                groups = align.get_matching_atoms(ref, mobile, strict=strict)
        else:
            with pytest.warns(SelectionWarning):
                with pytest.raises(SelectionError):
                    groups = align.get_matching_atoms(ref, mobile, strict=strict)

    @pytest.mark.parametrize("strict", (True, False))
    def test_nomatch_residues_raise_empty(self, universe, reference_small,
                                          strict, selection="protein and backbone"):
        # one atom less and all residues different: will currently create
        # empty selections with strict=False, see also
        # https://gist.github.com/orbeckst/2686badcd15031e6c946baf9164a683d
        ref = reference_small.select_atoms(selection)
        mobile = universe.select_atoms(selection)
        if strict:
            with pytest.raises(SelectionError):
                groups = align.get_matching_atoms(ref, mobile, strict=strict)
        else:
            with pytest.warns(SelectionWarning):
                with pytest.raises(SelectionError):
                    groups = align.get_matching_atoms(ref, mobile, strict=strict)

    def test_toggle_atom_mismatch_default_error(self, universe, reference):
        selection = ('resname ALA and name CA', 'resname ALA and name O')
        with pytest.raises(SelectionError):
            rmsd = align.alignto(universe, reference, select=selection)

    def test_toggle_atom_mismatch_kwarg_error(self, universe, reference):
        selection = ('resname ALA and name CA', 'resname ALA and name O')
        with pytest.raises(SelectionError):
            rmsd = align.alignto(universe, reference, select=selection, match_atoms=True)

    def test_toggle_atom_nomatch(self, universe, reference):
        selection = ('resname ALA and name CA', 'resname ALA and name O')
        rmsd = align.alignto(universe, reference, select=selection, match_atoms=False)
        assert rmsd[0] > 0.01

    def test_toggle_atom_nomatch_mismatch_atoms(self, universe, reference):
        # mismatching number of atoms, but same number of residues
        u = universe.select_atoms('resname ALA and name CA')
        u += universe.select_atoms('resname ALA and name O')[-1]
        ref = reference.select_atoms('resname ALA and name CA')
        with pytest.raises(SelectionError):
            align.alignto(u, ref, select='all', match_atoms=False)

    @pytest.mark.parametrize('subselection, expectation', [
        ('resname ALA and name CA', does_not_raise()),
        (mda.Universe(PSF, DCD).select_atoms('resname ALA and name CA'), does_not_raise()),
        (1234, pytest.raises(TypeError)),
    ])
    def test_subselection_alignto(self, universe, reference, subselection, expectation):

        with expectation:
            rmsd = align.alignto(universe, reference, subselection=subselection)
            assert_almost_equal(rmsd[1], 0.0, decimal=9)

    def test_no_atom_masses(self, universe):
        #if no masses are present
        u = mda.Universe.empty(6, 2, atom_resindex=[0, 0, 0, 1, 1, 1], trajectory=True)
        with pytest.warns(SelectionWarning):
            align.get_matching_atoms(u.atoms, u.atoms)

    def test_one_universe_has_masses(self, universe):
        u = mda.Universe.empty(6, 2, atom_resindex=[0, 0, 0, 1, 1, 1], trajectory=True)
        ref = mda.Universe.empty(6, 2, atom_resindex=[0, 0, 0, 1, 1, 1], trajectory=True)
        ref.add_TopologyAttr('masses')
        with pytest.warns(SelectionWarning):
            align.get_matching_atoms(u.atoms, ref.atoms)

class TestAlign(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PSF, DCD)

    @staticmethod
    @pytest.fixture()
    def reference():
        return mda.Universe(PSF, DCD)

    def test_rmsd(self, universe, reference):
        universe.trajectory[0]  # ensure first frame
        bb = universe.select_atoms('backbone')
        first_frame = bb.positions
        universe.trajectory[-1]
        last_frame = bb.positions
        assert_almost_equal(rms.rmsd(first_frame, first_frame), 0.0, 5,
                            err_msg="error: rmsd(X,X) should be 0")
        # rmsd(A,B) = rmsd(B,A) should be exact but spurious failures in the
        # 9th decimal have been observed (see Issue 57 comment #1) so we relax
        # the test to 6 decimals.
        rmsd = rms.rmsd(first_frame, last_frame, superposition=True)
        assert_almost_equal(
            rms.rmsd(last_frame, first_frame, superposition=True), rmsd, 6,
            err_msg="error: rmsd() is not symmetric")
        assert_almost_equal(rmsd, 6.820321761927005, 5,
                            err_msg="RMSD calculation between 1st and last AdK frame gave wrong answer")
        # test masses as weights
        last_atoms_weight = universe.atoms.masses
        A = universe.trajectory[0]
        B = reference.trajectory[-1]
        rmsd = align.alignto(universe, reference, weights='mass')
        rmsd_sup_weight = rms.rmsd(A, B, weights=last_atoms_weight, center=True,
                                   superposition=True)
        assert_almost_equal(rmsd[1], rmsd_sup_weight, 6)

    def test_rmsd_custom_mass_weights(self, universe, reference):
        last_atoms_weight = universe.atoms.masses
        A = universe.trajectory[0]
        B = reference.trajectory[-1]
        rmsd = align.alignto(universe, reference,
                             weights=reference.atoms.masses)
        rmsd_sup_weight = rms.rmsd(A, B, weights=last_atoms_weight, center=True,
                                   superposition=True)
        assert_almost_equal(rmsd[1], rmsd_sup_weight, 6)

    def test_rmsd_custom_weights(self, universe, reference):
        weights = np.zeros(universe.atoms.n_atoms)
        ca = universe.select_atoms('name CA')
        weights[ca.indices] = 1
        rmsd = align.alignto(universe, reference, select='name CA')
        rmsd_weights = align.alignto(universe, reference, weights=weights)
        assert_almost_equal(rmsd[1], rmsd_weights[1], 6)

    def test_AlignTraj_outfile_default(self, universe, reference, tmpdir):
        with tmpdir.as_cwd():
            reference.trajectory[-1]
            x = align.AlignTraj(universe, reference)
            try:
                assert os.path.basename(x.filename) == 'rmsfit_adk_dims.dcd'
            finally:
                x._writer.close()

    def test_AlignTraj_outfile_default_exists(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join('align_test.dcd'))
        align.AlignTraj(universe, reference, filename=outfile).run()
        fitted = mda.Universe(PSF, outfile)

        # ensure default file exists
        with mda.Writer(str(tmpdir.join("rmsfit_align_test.dcd")),
                        n_atoms=fitted.atoms.n_atoms) as w:
            w.write(fitted.atoms)

        with tmpdir.as_cwd():
            align.AlignTraj(fitted, reference)

            # we are careful now. The default does nothing
            with pytest.raises(IOError):
                align.AlignTraj(fitted, reference, force=False)

    def test_AlignTraj_step_works(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join('align_test.dcd'))
        # this shouldn't throw an exception
        align.AlignTraj(universe, reference, filename=outfile).run(step=10)

    def test_AlignTraj_deprecated_attribute(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join('align_test.dcd'))
        x = align.AlignTraj(universe, reference, filename=outfile).run(stop=2)

        wmsg = "The `rmsd` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(x.rmsd, x.results.rmsd)

    def test_AlignTraj(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join('align_test.dcd'))
        x = align.AlignTraj(universe, reference, filename=outfile).run()
        fitted = mda.Universe(PSF, outfile)

        assert_almost_equal(x.results.rmsd[0], 6.9290, decimal=3)
        assert_almost_equal(x.results.rmsd[-1], 5.2797e-07, decimal=3)

        # RMSD against the reference frame
        # calculated on Mac OS X x86 with MDA 0.7.2 r689
        # VMD: 6.9378711
        self._assert_rmsd(reference, fitted, 0, 6.929083044751061)
        self._assert_rmsd(reference, fitted, -1, 0.0)

    def test_AlignTraj_weighted(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join('align_test.dcd'))
        x = align.AlignTraj(universe, reference,
                            filename=outfile, weights='mass').run()
        fitted = mda.Universe(PSF, outfile)
        assert_almost_equal(x.results.rmsd[0], 0, decimal=3)
        assert_almost_equal(x.results.rmsd[-1], 6.9033, decimal=3)

        self._assert_rmsd(reference, fitted, 0, 0.0,
                          weights=universe.atoms.masses)
        self._assert_rmsd(reference, fitted, -1, 6.929083032629219,
                          weights=universe.atoms.masses)

    def test_AlignTraj_custom_weights(self, universe, reference, tmpdir):
        weights = np.zeros(universe.atoms.n_atoms)
        ca = universe.select_atoms('name CA')
        weights[ca.indices] = 1

        outfile = str(tmpdir.join('align_test.dcd'))

        x = align.AlignTraj(universe, reference,
                            filename=outfile, select='name CA').run()
        x_weights = align.AlignTraj(universe, reference,
                                    filename=outfile, weights=weights).run()

        assert_array_almost_equal(x.results.rmsd, x_weights.results.rmsd)

    def test_AlignTraj_custom_mass_weights(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join('align_test.dcd'))
        x = align.AlignTraj(universe, reference,
                            filename=outfile,
                            weights=reference.atoms.masses).run()
        fitted = mda.Universe(PSF, outfile)
        assert_almost_equal(x.results.rmsd[0], 0, decimal=3)
        assert_almost_equal(x.results.rmsd[-1], 6.9033, decimal=3)

        self._assert_rmsd(reference, fitted, 0, 0.0,
                          weights=universe.atoms.masses)
        self._assert_rmsd(reference, fitted, -1, 6.929083032629219,
                          weights=universe.atoms.masses)

    def test_AlignTraj_partial_fit(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join('align_test.dcd'))
        # fitting on a partial selection should still write the whole topology
        align.AlignTraj(universe, reference, select='resid 1-20',
                        filename=outfile, weights='mass').run()
        mda.Universe(PSF, outfile)

    def test_AlignTraj_in_memory(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join('align_test.dcd'))
        reference.trajectory[-1]
        x = align.AlignTraj(universe, reference, filename=outfile,
                            in_memory=True).run()
        assert x.filename is None
        assert_almost_equal(x.results.rmsd[0], 6.9290, decimal=3)
        assert_almost_equal(x.results.rmsd[-1], 5.2797e-07, decimal=3)

        # check in memory trajectory
        self._assert_rmsd(reference, universe, 0, 6.929083044751061)
        self._assert_rmsd(reference, universe, -1, 0.0)

    def _assert_rmsd(self, reference, fitted, frame, desired, weights=None):
        fitted.trajectory[frame]
        rmsd = rms.rmsd(reference.atoms.positions, fitted.atoms.positions,
                        superposition=True)
        assert_almost_equal(rmsd, desired, decimal=5,
                            err_msg="frame {0:d} of fit does not have "
                                    "expected RMSD".format(frame))

    def test_alignto_checks_selections(self, universe, reference):
        """Testing that alignto() fails if selections do not
        match (Issue 143)"""
        u = universe

        def different_size():
            a = u.atoms[10:100]
            b = u.atoms[10:101]
            return align.alignto(a, b)

        with pytest.raises(SelectionError):
            different_size()

        def different_atoms():
            a = u.atoms[10:20]
            b = u.atoms[10:17] + u.atoms[18:21]
            return align.alignto(a, b)

        with pytest.raises(SelectionError):
            different_atoms()

    def test_alignto_partial_universe(self, universe, reference):
        u_bound = mda.Universe(ALIGN_BOUND)
        u_free = mda.Universe(ALIGN_UNBOUND)
        selection = 'segid B'

        segB_bound = u_bound.select_atoms(selection)
        segB_free = u_free.select_atoms(selection)
        segB_free.translate(segB_bound.centroid() - segB_free.centroid())

        align.alignto(u_free, u_bound, select=selection)
        assert_array_almost_equal(segB_bound.positions, segB_free.positions,
                                  decimal=3)


def _get_aligned_average_positions(ref_files, ref, select="all", **kwargs):
    u = mda.Universe(*ref_files, in_memory=True)
    prealigner = align.AlignTraj(u, ref, select=select, **kwargs).run()
    ag = u.select_atoms(select)
    reference_coordinates = u.trajectory.timeseries(asel=ag).mean(axis=1)
    rmsd = sum(prealigner.results.rmsd/len(u.trajectory))
    return reference_coordinates, rmsd

class TestAverageStructure(object):

    ref_files = (PSF, DCD)

    @pytest.fixture
    def universe(self):
        return mda.Universe(*self.ref_files)

    @pytest.fixture
    def reference(self):
        return mda.Universe(PSF, CRD)

    def test_average_structure_deprecated_attrs(self, universe, reference):
        # Issue #3278 - remove in MDAnalysis 3.0.0
        avg = align.AverageStructure(universe, reference).run(stop=2)

        wmsg = "The `universe` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(avg.universe.atoms.positions,
                         avg.results.universe.atoms.positions)

        wmsg = "The `positions` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(avg.positions, avg.results.positions)

        wmsg = "The `rmsd` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert avg.rmsd == avg.results.rmsd

    def test_average_structure(self, universe, reference):
        ref, rmsd = _get_aligned_average_positions(self.ref_files, reference)
        avg = align.AverageStructure(universe, reference).run()
        assert_almost_equal(avg.results.universe.atoms.positions, ref,
                            decimal=4)
        assert_almost_equal(avg.results.rmsd, rmsd)

    def test_average_structure_mass_weighted(self, universe, reference):
        ref, rmsd = _get_aligned_average_positions(self.ref_files, reference, weights='mass')
        avg = align.AverageStructure(universe, reference, weights='mass').run()
        assert_almost_equal(avg.results.universe.atoms.positions, ref,
                            decimal=4)
        assert_almost_equal(avg.results.rmsd, rmsd)

    def test_average_structure_select(self, universe, reference):
        select = 'protein and name CA and resid 3-5'
        ref, rmsd = _get_aligned_average_positions(self.ref_files, reference, select=select)
        avg = align.AverageStructure(universe, reference, select=select).run()
        assert_almost_equal(avg.results.universe.atoms.positions, ref,
                            decimal=4)
        assert_almost_equal(avg.results.rmsd, rmsd)

    def test_average_structure_no_ref(self, universe):
        ref, rmsd = _get_aligned_average_positions(self.ref_files, universe)
        avg = align.AverageStructure(universe).run()
        assert_almost_equal(avg.results.universe.atoms.positions, ref,
                            decimal=4)
        assert_almost_equal(avg.results.rmsd, rmsd)

    def test_average_structure_no_msf(self, universe):
        avg = align.AverageStructure(universe).run()
        assert not hasattr(avg, 'msf')

    def test_mismatch_atoms(self, universe):
        u = mda.Merge(universe.atoms[:10])
        with pytest.raises(SelectionError):
            align.AverageStructure(universe, u)

    def test_average_structure_ref_frame(self, universe):
        ref_frame = 3
        u = mda.Merge(universe.atoms)

        # change to ref_frame
        universe.trajectory[ref_frame]
        u.load_new(universe.atoms.positions)

        # back to start
        universe.trajectory[0]
        ref, rmsd = _get_aligned_average_positions(self.ref_files, u)
        avg = align.AverageStructure(universe, ref_frame=ref_frame).run()
        assert_almost_equal(avg.results.universe.atoms.positions, ref,
                            decimal=4)
        assert_almost_equal(avg.results.rmsd, rmsd)

    def test_average_structure_in_memory(self, universe):
        avg = align.AverageStructure(universe, in_memory=True).run()
        reference_coordinates = universe.trajectory.timeseries().mean(axis=1)
        assert_almost_equal(avg.results.universe.atoms.positions,
                            reference_coordinates, decimal=4)
        assert avg.filename is None


class TestAlignmentProcessing(object):
    seq = FASTA
    error_msg = "selection string has unexpected length"

    def test_fasta2select_aligned(self):
        """test align.fasta2select() on aligned FASTA (Issue 112)"""
        sel = align.fasta2select(self.seq, is_aligned=True)
        # length of the output strings, not residues or anything real...
        assert len(sel['reference']) == 30623, self.error_msg
        assert len(sel['mobile']) == 30623, self.error_msg

    @pytest.mark.skipif(executable_not_found("clustalw2"),
                        reason="Test skipped because clustalw2 executable not found")
    def test_fasta2select_file(self, tmpdir):
        """test align.fasta2select() on a non-aligned FASTA with default
        filenames"""
        with tmpdir.as_cwd():
            sel = align.fasta2select(self.seq, is_aligned=False,
                                     alnfilename=None, treefilename=None)
            assert len(sel['reference']) == 23080, self.error_msg
            assert len(sel['mobile']) == 23090, self.error_msg

    @pytest.mark.skipif(executable_not_found("clustalw2"),
                        reason="Test skipped because clustalw2 executable not found")
    def test_fasta2select_ClustalW(self, tmpdir):
        """MDAnalysis.analysis.align: test fasta2select() with ClustalW
        (Issue 113)"""
        alnfile = str(tmpdir.join('alignmentprocessing.aln'))
        treefile = str(tmpdir.join('alignmentprocessing.dnd'))
        sel = align.fasta2select(self.seq, is_aligned=False,
                                 alnfilename=alnfile, treefilename=treefile)
        # numbers computed from alignment with clustalw 2.1 on Mac OS X
        # [orbeckst] length of the output strings, not residues or anything
        # real...
        assert len(sel['reference']) == 23080, self.error_msg
        assert len(sel['mobile']) == 23090, self.error_msg

    def test_fasta2select_resids(self, tmpdir):
        """test align.fasta2select() when resids provided (Issue #3124)"""
        resids = [x for x in range(705)]
        sel = align.fasta2select(self.seq, is_aligned=True,
                                 ref_resids=resids, target_resids=resids)
        # length of the output strings, not residues or anything real...
        assert len(sel['reference']) == 30621, self.error_msg
        assert len(sel['mobile']) == 30621, self.error_msg

def test_sequence_alignment():
    u = mda.Universe(PSF)
    reference = u.atoms
    mobile = u.select_atoms("resid 122-159")
    aln = align.sequence_alignment(mobile, reference)

    assert len(aln) == 5, "return value has wrong tuple size"

    seqA, seqB, score, begin, end = aln
    assert_equal(seqA, reference.residues.sequence(format="string"),
                 err_msg="reference sequence mismatch")
    assert mobile.residues.sequence(
        format="string") in seqB, "mobile sequence mismatch"
    assert_almost_equal(score, 54.6)
    assert_array_equal([begin, end], [0, reference.n_residues])


def test_alignto_reorder_atomgroups():
    # Issue 2977
    u = mda.Universe(PDB_helix)
    mobile = u.atoms[:4]
    ref = u.atoms[[3, 2, 1, 0]]
    rmsd = align.alignto(mobile, ref, select='bynum 1-4')
    assert_allclose(rmsd, (0.0, 0.0))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import pytest

from MDAnalysisTests.datafiles import PSF_TRICLINIC, DCD_TRICLINIC
from MDAnalysis.analysis.dielectric import DielectricConstant
from MDAnalysis.exceptions import NoDataError
from numpy.testing import assert_allclose


class TestDielectric(object):
    @pytest.fixture()
    def ag(self):
        u = mda.Universe(PSF_TRICLINIC, DCD_TRICLINIC, in_memory=True)
        return u.atoms

    def test_broken_molecules(self, ag):
        # cut molecules apart
        ag.universe.transfer_to_memory()
        for ts in ag.universe.trajectory:
            ag.wrap()

        eps = DielectricConstant(ag, make_whole=False).run()
        assert_allclose(eps.results['eps_mean'], 721.711, rtol=1e-03)

    def test_broken_repaired_molecules(self, ag):
        # cut molecules apart
        ag.universe.transfer_to_memory()
        for ts in ag.universe.trajectory:
            ag.wrap()

        eps = DielectricConstant(ag, make_whole=True).run()
        assert_allclose(eps.results['eps_mean'], 5.088, rtol=1e-03)

    def test_temperature(self, ag):
        eps = DielectricConstant(ag, temperature=100).run()
        assert_allclose(eps.results['eps_mean'], 9.621, rtol=1e-03)

    def test_non_charges(self):
        u = mda.Universe(DCD_TRICLINIC)
        with pytest.raises(NoDataError,
                           match="No charges defined given atomgroup."):
            DielectricConstant(u.atoms).run()

    def test_non_neutral(self, ag):
        with pytest.raises(NotImplementedError,
                           match="Analysis for non-neutral systems or"):
            DielectricConstant(ag[:-1]).run()

    def test_free_charges(self, ag):
        ag.fragments[0].charges += 1
        ag.fragments[1].charges -= 1

        with pytest.raises(NotImplementedError):
            DielectricConstant(ag).run()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


from MDAnalysis.analysis.msd import EinsteinMSD as MSD
import MDAnalysis as mda

from numpy.testing import (assert_almost_equal, assert_equal)
import numpy as np

from MDAnalysisTests.datafiles import PSF, DCD, RANDOM_WALK, RANDOM_WALK_TOPO
from MDAnalysisTests.util import block_import, import_not_available

import pytest


@pytest.fixture(scope='module')
def SELECTION():
    selection = 'backbone and name CA and resid 1-10'
    return selection


@pytest.fixture(scope='module')
def u():
    return mda.Universe(PSF, DCD)


@pytest.fixture(scope='module')
def NSTEP():
    nstep = 5000
    return nstep


@pytest.fixture(scope='module')
def random_walk_u():
    # 100x100
    return mda.Universe(RANDOM_WALK_TOPO, RANDOM_WALK)


@pytest.fixture(scope='module')
def msd(u, SELECTION):
    # non fft msd
    m = MSD(u, SELECTION, msd_type='xyz', fft=False)
    m.run()
    return m


@pytest.fixture(scope='module')
def step_traj(NSTEP):  # constant velocity
    x = np.arange(NSTEP)
    traj = np.vstack([x, x, x]).T
    traj_reshape = traj.reshape([NSTEP, 1, 3])
    u = mda.Universe.empty(1)
    u.load_new(traj_reshape)
    return u


@block_import('tidynamics')
def test_notidynamics(u, SELECTION):
    with pytest.raises(ImportError, match="tidynamics was not found"):
        u = mda.Universe(PSF, DCD)
        msd = MSD(u, SELECTION)
        msd.run()


def characteristic_poly(n, d):
    # polynomial that describes unit step traj MSD
    x = np.arange(0, n)
    y = d*x*x
    return y


class TestMSDSimple(object):

    def test_selection_works(self, msd):
        # test some basic size and shape things
        assert_equal(msd.n_particles, 10)

    def test_ag_accepted(self, u):
        ag = u.select_atoms("resid 1")
        m = MSD(ag, msd_type='xyz', fft=False)

    def test_updating_ag_rejected(self, u):
        updating_ag = u.select_atoms("around 3.5 resid 1", updating=True)
        errmsg = "UpdatingAtomGroups are not valid"
        with pytest.raises(TypeError, match=errmsg):
            m = MSD(updating_ag, msd_type='xyz', fft=False)

    @pytest.mark.parametrize('msdtype', ['foo', 'bar', 'yx', 'zyx'])
    def test_msdtype_error(self, u, SELECTION, msdtype):
        errmsg = f"invalid msd_type: {msdtype}"
        with pytest.raises(ValueError, match=errmsg):
            m = MSD(u, SELECTION, msd_type=msdtype)

    @pytest.mark.parametrize("dim, dim_factor", [
        ('xyz', 3), ('xy', 2), ('xz', 2), ('yz', 2), ('x', 1), ('y', 1),
        ('z', 1)
    ])
    def test_simple_step_traj_all_dims(self, step_traj, NSTEP, dim,
                                       dim_factor):
        # testing the "simple" algorithm on constant velocity trajectory
        # should fit the polynomial y=dim_factor*x**2
        m_simple = MSD(step_traj, 'all', msd_type=dim, fft=False)
        m_simple.run()
        poly = characteristic_poly(NSTEP, dim_factor)
        assert_almost_equal(m_simple.results.timeseries, poly, decimal=4)

    @pytest.mark.parametrize("dim, dim_factor", [
        ('xyz', 3), ('xy', 2), ('xz', 2), ('yz', 2), ('x', 1), ('y', 1),
        ('z', 1)
    ])
    def test_simple_start_stop_step_all_dims(self, step_traj, NSTEP, dim,
                                             dim_factor):
        # testing the "simple" algorithm on constant velocity trajectory
        # test start stop step is working correctly
        m_simple = MSD(step_traj, 'all', msd_type=dim, fft=False)
        m_simple.run(start=10, stop=1000, step=10)
        poly = characteristic_poly(NSTEP, dim_factor)
        # polynomial must take offset start into account
        assert_almost_equal(m_simple.results.timeseries, poly[0:990:10],
                            decimal=4)

    def test_random_walk_u_simple(self, random_walk_u):
        # regress against random_walk test data
        msd_rw = MSD(random_walk_u, 'all', msd_type='xyz', fft=False)
        msd_rw.run()
        norm = np.linalg.norm(msd_rw.results.timeseries)
        val = 3932.39927487146
        assert_almost_equal(norm, val, decimal=5)


@pytest.mark.skipif(import_not_available("tidynamics"),
                    reason="Test skipped because tidynamics not found")
class TestMSDFFT(object):

    @pytest.fixture(scope='class')
    def msd_fft(self, u, SELECTION):
        # fft msd
        m = MSD(u, SELECTION, msd_type='xyz', fft=True)
        m.run()
        return m

    def test_fft_vs_simple_default(self, msd, msd_fft):
        # testing on the  PSF, DCD trajectory
        timeseries_simple = msd.results.timeseries
        timeseries_fft = msd_fft.results.timeseries
        assert_almost_equal(timeseries_simple, timeseries_fft, decimal=4)

    def test_fft_vs_simple_default_per_particle(self, msd, msd_fft):
        # check fft and simple give same result per particle
        per_particle_simple = msd.results.msds_by_particle
        per_particle_fft = msd_fft.results.msds_by_particle
        assert_almost_equal(per_particle_simple, per_particle_fft, decimal=4)

    @pytest.mark.parametrize("dim", ['xyz', 'xy', 'xz', 'yz', 'x', 'y', 'z'])
    def test_fft_vs_simple_all_dims(self, u, SELECTION, dim):
        # check fft and simple give same result for each dimensionality
        m_simple = MSD(u, SELECTION, msd_type=dim, fft=False)
        m_simple.run()
        timeseries_simple = m_simple.results.timeseries
        m_fft = MSD(u, SELECTION, msd_type=dim, fft=True)
        m_fft.run()
        timeseries_fft = m_fft.results.timeseries
        assert_almost_equal(timeseries_simple, timeseries_fft, decimal=4)

    @pytest.mark.parametrize("dim", ['xyz', 'xy', 'xz', 'yz', 'x', 'y', 'z'])
    def test_fft_vs_simple_all_dims_per_particle(self, u, SELECTION, dim):
        # check fft and simple give same result for each particle in each
        # dimension
        m_simple = MSD(u, SELECTION, msd_type=dim, fft=False)
        m_simple.run()
        per_particle_simple = m_simple.results.msds_by_particle
        m_fft = MSD(u, SELECTION, msd_type=dim, fft=True)
        m_fft.run()
        per_particle_fft = m_fft.results.msds_by_particle
        assert_almost_equal(per_particle_simple, per_particle_fft, decimal=4)

    @pytest.mark.parametrize("dim, dim_factor", [
        ('xyz', 3), ('xy', 2), ('xz', 2), ('yz', 2), ('x', 1), ('y', 1),
        ('z', 1)
    ])
    def test_fft_step_traj_all_dims(self, step_traj, NSTEP, dim, dim_factor):
        # testing the fft algorithm on constant velocity trajectory
        # this should fit the polynomial y=dim_factor*x**2
        # fft based tests require a slight decrease in expected prescision
        # primarily due to roundoff in fft(ifft()) calls.
        # relative accuracy expected to be around ~1e-12
        m_simple = MSD(step_traj, 'all', msd_type=dim, fft=True)
        m_simple.run()
        poly = characteristic_poly(NSTEP, dim_factor)
        # this was relaxed from decimal=4 for numpy=1.13 test
        assert_almost_equal(m_simple.results.timeseries, poly, decimal=3)

    @pytest.mark.parametrize("dim, dim_factor", [(
        'xyz', 3), ('xy', 2), ('xz', 2), ('yz', 2), ('x', 1), ('y', 1),
        ('z', 1)
    ])
    def test_fft_start_stop_step_all_dims(self, step_traj, NSTEP, dim,
                                          dim_factor):
        # testing the fft algorithm on constant velocity trajectory
        # test start stop step is working correctly
        m_simple = MSD(step_traj, 'all', msd_type=dim, fft=True)
        m_simple.run(start=10, stop=1000, step=10)
        poly = characteristic_poly(NSTEP, dim_factor)
        # polynomial must take offset start into account
        assert_almost_equal(m_simple.results.timeseries, poly[0:990:10],
                            decimal=3)

    def test_random_walk_u_fft(self, random_walk_u):
        # regress against random_walk test data
        msd_rw = MSD(random_walk_u, 'all', msd_type='xyz', fft=True)
        msd_rw.run()
        norm = np.linalg.norm(msd_rw.results.timeseries)
        val = 3932.39927487146
        assert_almost_equal(norm, val, decimal=5)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_equal
import pytest

import MDAnalysis as mda

from MDAnalysisTests.datafiles import PSF, DCD


class TestSequence(object):
    # all tests are done with the AdK system (PSF and DCD) sequence:
    # http://www.uniprot.org/uniprot/P69441.fasta
    # >sp|P69441|KAD_ECOLI Adenylate kinase OS=Escherichia coli (strain K12) GN=adk PE=1 SV=1
    ref_adk_sequence = (
        "MRIILLGAPGAGKGTQAQFIMEKYGIPQISTGDMLRAAVKSGSELGKQAKDIMDAGKLVT"
        "DELVIALVKERIAQEDCRNGFLLDGFPRTIPQADAMKEAGINVDYVLEFDVPDELIVDRI"
        "VGRRVHAPSGRVYHVKFNPPKVEGKDDVTGEELTTRKDDQEETVRKRLVEYHQMTAPLIG"
        "YYSKEAEAGNTKYAKVDGTKPVAEVRADLEKILG"
    )

    @pytest.fixture()
    def u(self):
        return mda.Universe(PSF, DCD)

    def test_string(self, u):
        p = u.select_atoms("protein")
        assert_equal(p.residues.sequence(format="string"),
                     self.ref_adk_sequence)

    def test_SeqRecord(self, u):
        p = u.select_atoms("protein")
        s = p.residues.sequence(format="SeqRecord",
                                id="P69441", name="KAD_ECOLI Adenylate kinase",
                                description="EcAdK from pdb 4AKE")
        assert_equal(s.id, "P69441")
        assert_equal(str(s.seq), self.ref_adk_sequence)

    def test_SeqRecord_default(self, u):
        p = u.select_atoms("protein")
        s = p.residues.sequence(id="P69441", name="KAD_ECOLI Adenylate kinase",
                                description="EcAdK from pdb 4AKE")
        assert_equal(s.id, "P69441")
        assert_equal(str(s.seq), self.ref_adk_sequence)

    def test_Seq(self, u):
        p = u.select_atoms("protein")
        s = p.residues.sequence(format="Seq")
        assert_equal(str(s), self.ref_adk_sequence)

    def test_nonIUPACresname_VE(self, u):
        """test_sequence_nonIUPACresname: non recognized amino acids raise
        ValueError"""
        # fake non-IUPAC residue name for this test
        residues = u.select_atoms("resname MET").residues
        residues.resnames = "MSE"

        def wrong_res():
            u.residues.sequence()

        with pytest.raises(ValueError):
            wrong_res()

    def test_format_TE(self, u):
        with pytest.raises(TypeError):
            u.residues.sequence(format='chicken')


class TestResidueGroup(object):
    # Legacy tests from before 363
    @pytest.fixture()
    def universe(self):
        return mda.Universe(PSF, DCD)

    @pytest.fixture()
    def rg(self, universe):
        return universe.residues

    def test_newResidueGroup(self, universe):
        """test that slicing a ResidueGroup returns a new ResidueGroup
        (Issue 135)"""
        rg = universe.atoms.residues
        newrg = rg[10:20:2]
        assert isinstance(newrg, mda.core.groups.ResidueGroup), \
            "Failed to make a new ResidueGroup: type mismatch"

    def test_n_atoms(self, rg):
        assert_equal(rg.n_atoms, 3341)

    def test_n_residues(self, rg):
        assert_equal(rg.n_residues, 214)

    def test_resids_dim(self, rg):
        assert_equal(len(rg.resids), len(rg))

    def test_resnums_dim(self, rg):
        assert_equal(len(rg.resnums), len(rg))

    def test_segids_dim(self, rg):
        assert_equal(len(rg.segids), len(rg))

    def test_len(self, rg):
        """testing that len(residuegroup) == residuegroup.n_residues"""
        assert_equal(len(rg), rg.n_residues,
                     "len and n_residues disagree")

    def test_set_resids(self, universe):
        rg = universe.select_atoms("bynum 12:42").residues
        resid = 999
        rg.resids = resid
        # check individual atoms
        for at in rg.atoms:
            assert_equal(at.resid, resid,
                         err_msg="failed to set_resid atoms 12:42 to same resid")
        # check residues
        assert_equal(rg.resids, resid * np.ones(rg.n_residues),
                     err_msg="failed to set_resid of residues belonging to "
                             "atoms 12:42 to same resid")

    def test_set_resids(self, universe):
        """test_set_resid: set ResidueGroup resids on a per-residue basis"""
        rg = universe.select_atoms("resid 10:18").residues
        resids = np.array(rg.resids) + 1000
        rg.resids = resids
        # check individual atoms
        for r, resid in zip(rg, resids):
            for at in r.atoms:
                assert_equal(at.resid, resid,
                             err_msg="failed to set_resid residues 10:18 to same "
                                     "resid in residue {0}\n"
                                     "(resids = {1}\nresidues = {2})".format(r,
                                                                             resids,
                                                                             rg))
        assert_equal(rg.resids, resids,
                     err_msg="failed to set_resid of residues belonging to "
                             "residues 10:18 to new resids")

    def test_set_resids_updates_self(self, universe):
        rg = universe.select_atoms("resid 10:18").residues
        resids = np.array(rg.resids) + 1000
        rg.resids = resids
        assert_equal(rg.resids, resids,
                     err_msg="old selection was not changed in place "
                             "after set_resid")

    def test_set_resnum_single(self, universe):
        rg = universe.residues[:3]
        new = 22
        rg.resnums = new

        assert_equal(all(rg.resnums == new), True)
        for r in rg:
            assert_equal(r.resnum, new)

    def test_set_resnum_many(self, universe):
        rg = universe.residues[:3]
        new = [22, 23, 24]
        rg.resnums = new

        assert_equal(all(rg.resnums == new), True)
        for r, v in zip(rg, new):
            assert_equal(r.resnum, v)

    def test_set_resnum_ValueError(self, universe):
        rg = universe.residues[:3]
        new = [22, 23, 24, 25]
        with pytest.raises(ValueError):
            setattr(rg, 'resnums', new)

    # INVALID: no `set_resnames` method; use `resnames` property directly
    @pytest.mark.skip
    def test_set_resname_single(self, universe):
        rg = universe.residues[:3]
        new = 'newname'

        rg.set_resnames(new)
        assert_equal(all(rg.resnames == new), True)
        for r in rg:
            assert_equal(r.name, new)

    # INVALID: no `set_resnames` method; use `resnames` property directly
    @pytest.mark.skip
    def test_set_resname_many(self, universe):
        rg = universe.residues[:3]
        new = ['a', 'b', 'c']
        rg.set_resnames(new)

        assert_equal(all(rg.resnames == new), True)
        for r, v in zip(rg, new):
            assert_equal(r.name, v)

    # INVALID: no `set_resnames` method; use `resnames` property directly
    @pytest.mark.skip
    def test_set_resname_ValueError(self, universe):
        rg = universe.residues[:3]
        new = ['a', 'b', 'c', 'd']

        with pytest.raises(ValueError):
            rg.set_resnames(new)

    # INVALID: no `set_resids` method; also, residues are not mergeable
    # by setting resids; resids are not necessarily unique; atoms must
    # have their resindex set to change residue membership
    @pytest.mark.skip
    def test_merge_residues(self, universe):
        rg = universe.select_atoms("resid 12:14").residues
        nres_old = universe.atoms.n_residues
        natoms_old = rg.n_atoms
        rg.set_resids(12)  # merge all into one with resid 12
        nres_new = universe.atoms.n_residues
        r_merged = universe.select_atoms("resid 12:14").residues
        natoms_new = universe.select_atoms("resid 12").n_atoms
        assert_equal(len(r_merged), 1, err_msg="set_resid failed to merge "
                                               "residues: merged = {0}".format(
            r_merged))
        assert_equal(nres_new, nres_old - 2,
                     err_msg="set_resid failed to merge residues: "
                             "merged = {0}".format(r_merged))
        assert_equal(natoms_new, natoms_old, err_msg="set_resid lost atoms "
                                                     "on merge".format(
            r_merged))

        assert_equal(universe.residues.n_residues,
                     universe.atoms.n_residues,
                     err_msg="Universe.residues and Universe.atoms.n_residues "
                             "do not agree after residue "
                             "merge.")

    # INVALID: no `set_masses` method; use `masses` property directly
    @pytest.mark.skip
    def test_set_masses(self, universe):
        rg = universe.select_atoms("bynum 12:42 and name H*").residues
        mass = 2.0
        rg.set_masses(mass)
        # check individual atoms
        assert_equal([a.mass for a in rg.atoms],
                     mass * np.ones(rg.n_atoms),
                     err_msg="failed to set_mass H* atoms in resid 12:42 to {0}".format(
                         mass))

    # VALID
    def test_atom_order(self, universe):
        assert_equal(universe.residues.atoms.indices,
                     sorted(universe.residues.atoms.indices))

    def test_get_next_residue(self, rg):
        unsorted_rep_res = rg[[0, 1, 8, 3, 4, 0, 3, 1, -1]]
        next_res = unsorted_rep_res._get_next_residues_by_resid()
        resids = list(unsorted_rep_res.resids+1)
        resids[-1] = None
        next_resids = [r.resid if r is not None else None for r in next_res]
        assert_equal(len(next_res), len(unsorted_rep_res))
        assert_equal(next_resids, resids)

    def test_get_prev_residue(self, rg):
        unsorted_rep_res = rg[[0, 1, 8, 3, 4, 0, 3, 1, -1]]
        prev_res = unsorted_rep_res._get_prev_residues_by_resid()
        resids = list(unsorted_rep_res.resids-1)
        resids[0] = resids[5] = None
        prev_resids = [r.resid if r is not None else None for r in prev_res]
        assert_equal(len(prev_res), len(unsorted_rep_res))
        assert_equal(prev_resids, resids)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import os

import MDAnalysisTests
import pytest

"""Relative imports are banned in unit testing modules (Issue #189), so run tests to enforce this policy."""

path_to_testing_modules = MDAnalysisTests.__path__[0]
# Exclusion path relative to MDAnalysisTests
exclusions = ['/plugins', '/data']


def is_excluded(path):
    leaf = path[len(path_to_testing_modules):]
    return leaf in exclusions


def get_file_paths():
    paths = []
    for dirpath, dirnames, files in os.walk(path_to_testing_modules):
        if is_excluded(dirpath):
            continue
        for f in filter(lambda x: x.endswith('.py'), files):
            fpath = os.path.join(dirpath, f)
            if is_excluded(fpath):
                continue
            paths.append(fpath)
    return paths


@pytest.mark.parametrize('testing_module', get_file_paths())
def test_relative_import(testing_module):
    with open(testing_module, 'r') as test_module_file_object:
        for lineno, line in enumerate(test_module_file_object, start=1):
            if 'from .' in line and 'import' in line \
                    and not 'test_imports' in testing_module:
                raise AssertionError(
                    "A relative import statement was found in "
                    "module {testing_module} at linenumber {lineno}.".format(**vars()))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from collections import OrderedDict
from MDAnalysis.coordinates.base import (
    Timestep,
    SingleFrameReaderBase,
    ReaderBase
)
from numpy.testing import assert_equal, assert_allclose

import pytest
"""
Isolate the API definitions of Readers independent of implementations
"""


class AmazingMultiFrameReader(ReaderBase):
    format = 'AmazingMulti'

    def __init__(self, filename, **kwargs):
        self.filename = filename
        self.n_frames = 10
        self.n_atoms = 10
        self._auxs = {}
        self._transformations = []
        # ts isn't a real timestep, but just an integer
        # whose value represents the frame number (0 based)
        self.ts = Timestep(self.n_atoms)
        self.ts.frame = -1
        self._read_next_timestep()

    def _read_next_timestep(self):
        self.ts.frame += 1
        if (self.ts.frame + 1) > self.n_frames:
            raise IOError
        else:
            return self.ts

    def _read_frame(self, frame):
        if frame < 0:
            frame = self.n_frames + frame
        if not (0 <= frame < self.n_frames):
            raise IOError
        self.ts.frame = frame

        return self.ts

    def _reopen(self):
        self.ts.frame = -1


class AmazingReader(SingleFrameReaderBase):
    format = 'Amazing'

    # have to hack this in to get the base class to "work"
    def _read_first_frame(self):
        self.n_atoms = 10
        self.ts = Timestep(self.n_atoms)
        self.ts.frame = 0


class _TestReader(object):
    __test__ = False
    """Basic API readers"""

    @pytest.fixture()
    def reader(self):
        return self.readerclass('test.txt')

    @pytest.fixture()
    def ts(self, reader):
        return reader.ts

    def test_required_attributes(self, reader):
        """Test that Reader has the required attributes"""
        for attr in ['filename', 'n_atoms', 'n_frames', 'ts',
                     'units', 'format']:
            assert_equal(hasattr(reader, attr), True,
                         "Missing attr: {0}".format(attr))

    def test_iter(self, reader):
        l = [ts for ts in reader]

        assert_equal(len(l), self.n_frames)

    def test_close(self):
        sfr = self.readerclass('text.txt')

        ret = sfr.close()
        # Check that method works?
        assert_equal(ret, None)

    def test_rewind(self, reader):
        ret = reader.rewind()

        assert_equal(ret, None)
        assert_equal(reader.ts.frame, 0)

    def test_context(self):
        with self.readerclass('text.txt') as sfr:
            l = sfr.ts.frame

        assert_equal(l, 0)

    def test_len(self, reader):
        l = len(reader)

        assert_equal(l, self.n_frames)

    def test_raises_StopIteration(self, reader):
        reader[-1]
        with pytest.raises(StopIteration):
            next(reader)


class _Multi(_TestReader):
    n_frames = 10
    n_atoms = 10
    readerclass = AmazingMultiFrameReader
    reference = np.arange(10)


class TestMultiFrameReader(_Multi):

    __test__ = True

    @pytest.mark.parametrize('start, stop, step', [
        (None, None, None),  # blank slice
        (None, 5, None),  # set end point
        (2, None, None),  # set start point
        (2, 5, None),  # start & end
        (None, None, 2),  # set skip
        (None, None, -1),  # backwards skip
        (None, -1, -1),
        (10, 0, -1),
        (0, 10, 1),
        (0, 10, 2),
        (None, 20, None),  # end beyond real end
        (None, 20, 2),  # with skip
        (0, 5, 2),
        (5, None, -1),
        (None, 5, -1),
        (100, 10, 1),
        (-10, None, 1),
        (100, None, -1),  # beyond real end
        (100, 5, -20),
        (5, 1, 1),  # Stop less than start
        (1, 5, -1),  # Stop less than start
        (-100, None, None),
        (100, None, None),  # Outside of range of trajectory
        (-2, 10, -2),
        (0, 0, 1),  # empty
        (10, 1, 2), # empty
    ])
    def test_slice(self, start, stop, step, reader):
        """Compare the slice applied to trajectory, to slice of list"""
        res = [ts.frame for ts in reader[start:stop:step]]
        ref = self.reference[start:stop:step]

        assert_equal(res, ref)

    def test_slice_VE_1(self, reader):
        def sl():
            return list(reader[::0])

        with pytest.raises(ValueError):
            sl()

    def test_slice_TE_1(self, reader):
        def sl():
            return list(reader[1.2:2.5:0.1])

        with pytest.raises(TypeError):
            sl()

    @pytest.mark.parametrize('slice_cls', [list, np.array])
    @pytest.mark.parametrize('sl', [
        [0, 1, 4, 5],
        [5, 1, 6, 2, 7, 3, 8],
        [0, 1, 1, 1, 0, 0, 2, 3, 4],
        [True, False, True, False, True, False, True, False, True, False],
        [True, True, False, False, True, True, False, True, False, True],
        [True, True, True, True, True, True, True, True, True, True],
        [False, False, False, False, False, False, False, False, False, False],
    ])
    def test_getitem(self, slice_cls, sl, reader):
        sl = slice_cls(sl)
        res = [ts.frame for ts in reader[sl]]

        sl = np.asarray(sl)
        ref = self.reference[sl]

        assert_equal(res, ref)

    @pytest.mark.parametrize('sl', [
        [0, 1, 2, 3],  # ordered list of indices without duplicates
        [1, 3, 4, 2, 9],  # disordered list of indices without duplicates
        [0, 1, 1, 2, 2, 2],  # ordered list with duplicates
        [-1, -2, 3, -1, 0],  # disordered list with duplicates
        [True, ] * 10,
        [False, ] * 10,
        [True, False, ] * 5,
        slice(None, None, None),
        slice(0, 10, 1),
        slice(None, None, -1),
        slice(10, 0, -1),
        slice(2, 7, 2),
        slice(7, 2, -2),
        slice(7, 2, 1),  # empty
        slice(0, 0, 1),  # empty
    ])
    def test_getitem_len(self, sl, reader):
        traj_iterable = reader[sl]
        if not isinstance(sl, slice):
            sl = np.array(sl)
        ref = self.reference[sl]
        assert len(traj_iterable) == len(ref)

    @pytest.mark.parametrize('iter_type', (list, np.array))
    def test_getitem_len_empty(self, reader, iter_type):
        # Indexing a numpy array with an empty array tends to break.
        traj_iterable = reader[iter_type([])]
        assert len(traj_iterable) == 0

    # All the sl1 slice must be 5 frames long so that the sl2 can be a mask
    @pytest.mark.parametrize('sl1', [
        [0, 1, 2, 3, 4],
        [1, 1, 1, 1, 1],
        [True, False, ] * 5,
        slice(None, None, 2),
        slice(None, None, -2),
    ])
    @pytest.mark.parametrize('sl2', [
        [0, -1, 2],
        [-1,-1, -1],
        [True, False, True, True, False],
        np.array([True, False, True, True, False]),
        slice(None, None, None),
        slice(None, 3, None),
        slice(4, 0, -1),
        slice(None, None, -1),
        slice(None, None, 2),
    ])
    def test_double_getitem(self, sl1, sl2, reader):
        traj_iterable = reader[sl1][sl2]
        # Old versions of numpy do not behave the same when indexing with a
        # list or with an array.
        if not isinstance(sl1, slice):
            sl1 = np.asarray(sl1)
        if not isinstance(sl2, slice):
            sl2 = np.asarray(sl2)
        print(sl1, sl2, type(sl1), type(sl2))
        ref = self.reference[sl1][sl2]
        res = [ts.frame for ts in traj_iterable]
        assert_equal(res, ref)
        assert len(traj_iterable) == len(ref)

    @pytest.mark.parametrize('sl1', [
        [0, 1, 2, 3, 4],
        [1, 1, 1, 1, 1],
        [True, False, ] * 5,
        slice(None, None, 2),
        slice(None, None, -2),
        slice(None, None, None),
    ])
    @pytest.mark.parametrize('idx2', [0, 2, 4, -1, -2, -4])
    def test_double_getitem_int(self, sl1, idx2, reader):
        ts = reader[sl1][idx2]
        # Old versions of numpy do not behave the same when indexing with a
        # list or with an array.
        if not isinstance(sl1, slice):
            sl1 = np.asarray(sl1)
        ref = self.reference[sl1][idx2]
        assert ts.frame == ref

    def test_list_TE(self, reader):
        def sl():
            return list(reader[[0, 'a', 5, 6]])

        with pytest.raises(TypeError):
            sl()

    def test_array_TE(self, reader):
        def sl():
            return list(reader[np.array([1.2, 3.4, 5.6])])

        with pytest.raises(TypeError):
            sl()

    @pytest.mark.parametrize('sl1', [
        [0, 1, 2, 3, 4],
        [1, 1, 1, 1, 1],
        [True, False, ] * 5,
        slice(None, None, 2),
        slice(None, None, -2),
    ])
    @pytest.mark.parametrize('idx2', [5, -6])
    def test_getitem_IE(self, sl1, idx2, reader):
        partial_reader = reader[sl1]
        with pytest.raises(IndexError):
            partial_reader[idx2]


class _Single(_TestReader):
    n_frames = 1
    n_atoms = 10
    readerclass = AmazingReader


class TestSingleFrameReader(_Single):
    __test__ = True
    def test_next(self, reader):
        with pytest.raises(StopIteration):
            reader.next()

    # Getitem tests
    # only 0 & -1 should work
    # others should get IndexError
    def _check_get_results(self, l, ts):
        assert_equal(len(l), 1)
        assert_equal(ts in l, True)

    def test_getitem(self, reader, ts):
        fr = [reader[0]]

        self._check_get_results(fr, ts)

    def test_getitem_2(self, reader, ts):
        fr = [reader[-1]]

        self._check_get_results(fr, ts)

    def test_getitem_IE(self, reader):
        with pytest.raises(IndexError):
            reader.__getitem__(1)

    def test_getitem_IE_2(self, reader):
        with pytest.raises(IndexError):
            reader.__getitem__(-2)

    # Slicing should still work!
    def test_slice_1(self, reader, ts):
        l = list(reader[::])
        self._check_get_results(l, ts)

    def test_slice_2(self, reader, ts):
        l = list(reader[::-1])
        self._check_get_results(l, ts)

    def test_reopen(self, reader, ts):
        reader._reopen()
        assert_equal(ts.frame, 0)

    def test_rewind(self, reader, ts):
        reader.rewind()
        assert_equal(ts.frame, 0)

    def test_read_frame(self, reader):
        with pytest.raises(IndexError):
            reader._read_frame(1)

    def test_iter_rewind(self, reader):
        # Issue #3423
        # positions should be zero to start with
        assert_allclose(reader.ts.positions, np.zeros((10, 3)))

        # modify positions in place
        reader.ts.positions = np.ones((10, 3))
        assert_allclose(reader.ts.positions, np.ones((10, 3)))

        # iterate the reader
        for ts in reader:
            assert_allclose(ts.positions, np.zeros((10, 3)))

        assert_allclose(reader.ts.positions, np.zeros((10, 3)))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- http://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
try:
    from numpy import shares_memory
except ImportError:
    shares_memory = False

from numpy.testing import assert_equal, assert_almost_equal
import pytest
import MDAnalysis as mda

from MDAnalysisTests.datafiles import (
    AUX_XVG,
    CRD,
    DCD,
    DMS,
    DLP_CONFIG,
    DLP_HISTORY,
    FHIAIMS,
    INPCRD,
    GMS_ASYMOPT,
    GRO,
    GSD,
    LAMMPSdata_mini,
    mol2_molecules,
    MMTF,
    NAMDBIN,
    NCDF,
    PDB_small,
    PDBQT_input,
    PQR,
    TRR,
    TRJ,
    TRZ,
    TXYZ,
    XTC,
    XPDB_small,
    XYZ_mini,
)
from MDAnalysis.coordinates.core import get_reader_for


@pytest.fixture(params=[
    # formatname, filename
    ('CRD', CRD, dict()),
    ('DATA', LAMMPSdata_mini, dict(n_atoms=1)),
    ('DCD', DCD, dict()),
    ('DMS', DMS, dict()),
    ('CONFIG', DLP_CONFIG, dict()),
    ('HISTORY', DLP_HISTORY, dict()),
    ('INPCRD', INPCRD, dict()),
    ('GMS', GMS_ASYMOPT, dict()),
    ('GRO', GRO, dict()),
    ('MMTF', MMTF, dict()),
    ('MOL2', mol2_molecules, dict()),
    ('PDB', PDB_small, dict()),
    ('PQR', PQR, dict()),
    ('PDBQT', PDBQT_input, dict()),
    ('TRR', TRR, dict()),
    ('TRZ', TRZ, dict(n_atoms=8184)),
    ('TRJ', TRJ, dict(n_atoms=252)),
    ('XTC', XTC, dict()),
    ('XPDB', XPDB_small, dict()),
    ('XYZ', XYZ_mini, dict()),
    ('NCDF', NCDF, dict()),
    ('memory', np.arange(60).reshape(2, 10, 3).astype(np.float64), dict()),
    ('CHAIN', [GRO, GRO, GRO], dict()),
    ('FHIAIMS', FHIAIMS, dict()),
    ('GSD', GSD, dict()),
    ('NAMDBIN', NAMDBIN, dict()),
    ('TXYZ', TXYZ, dict()),
])
def ref_reader(request):
    fmt_name, filename, extras = request.param

    r = get_reader_for(filename, format=fmt_name)(filename, **extras)
    try:
        yield r
    finally:
        # make sure file handle is closed afterwards
        r.close()


@pytest.fixture(params=[
    # formatname, filename
    ('CRD', CRD, dict()),
    ('DATA', LAMMPSdata_mini, dict(n_atoms=1)),
    ('DCD', DCD, dict()),
    ('DMS', DMS, dict()),
    ('CONFIG', DLP_CONFIG, dict()),
    ('HISTORY', DLP_HISTORY, dict()),
    ('INPCRD', INPCRD, dict()),
    ('GMS', GMS_ASYMOPT, dict()),
    ('GRO', GRO, dict()),
    ('MMTF', MMTF, dict()),
    ('MOL2', mol2_molecules, dict()),
    ('PDB', PDB_small, dict()),
    ('PQR', PQR, dict()),
    ('PDBQT', PDBQT_input, dict()),
    ('TRR', TRR, dict()),
    ('TRZ', TRZ, dict(n_atoms=8184)),
    ('TRJ', TRJ, dict(n_atoms=252)),
    ('XTC', XTC, dict()),
    ('XPDB', XPDB_small, dict()),
    ('XYZ', XYZ_mini, dict()),
    ('NCDF', NCDF, dict(mmap=False)),
    ('memory', np.arange(60).reshape(2, 10, 3).astype(np.float64), dict()),
    ('CHAIN', [GRO, GRO, GRO], dict()),
    ('FHIAIMS', FHIAIMS, dict()),
    ('GSD', GSD, dict()),
    ('NAMDBIN', NAMDBIN, dict()),
    ('TXYZ', TXYZ, dict()),
])
def ref_reader_extra_args(request):
    fmt_name, filename, extras = request.param

    r = get_reader_for(filename, format=fmt_name)(
            filename, convert_units=False, dt=2, time_offset=10,
            foo="bar", **extras)
    try:
        yield r
    finally:
        # make sure file handle is closed afterwards
        r.close()


@pytest.fixture()
def original_and_copy(ref_reader):
    new = ref_reader.copy()
    try:
        yield ref_reader, new
    finally:
        new.close()


@pytest.fixture()
def original_and_copy_extra_args(ref_reader_extra_args):
    new = ref_reader_extra_args.copy()
    try:
        yield ref_reader_extra_args, new
    finally:
        new.close()


def test_reader_n_atoms(original_and_copy):
    original, copy = original_and_copy
    assert original.n_atoms == copy.n_atoms


def test_reader_filename(original_and_copy):
    original, copy = original_and_copy
    assert original.filename == copy.filename


def test_reader_copied_extra_attributes(original_and_copy_extra_args):
    # Issue #3664
    original, copy = original_and_copy_extra_args

    # memory and chain readers subclass protoreader directly and
    # therefore don't have convert_units or _ts_kwargs
    if original.__class__.__bases__[0].__name__ != "ProtoReader":
        assert original.format not in ('MEMORY', 'CHAIN')
        assert original.convert_units is False
        assert copy.convert_units is False
        assert original._ts_kwargs['time_offset'] == 10
        assert copy._ts_kwargs['time_offset'] == 10
        assert original._ts_kwargs['dt'] == 2
        assert copy._ts_kwargs['dt'] == 2

    assert original.ts.data['time_offset'] == 10
    assert copy.ts.data['time_offset'] == 10

    # Issue #3689 XTC and XDR overwrite `dt`
    if original.format not in ('XTC', 'TRR'):
        assert original.ts.data['dt'] == 2
        assert copy.ts.data['dt'] == 2

    assert copy._kwargs['foo'] == 'bar'

    # checking that non-base attributes are also copied (netcdf reader)
    if hasattr(original, "_mmap"):
        assert original._mmap is False
        assert copy._mmap is False


def test_reader_independent_iteration(original_and_copy):
    # check that the two Readers iterate independently
    original, copy = original_and_copy
    if len(original) < 2:
        pytest.skip('Single frame reader')
    # initially at same frame
    assert original.ts.frame == copy.ts.frame

    copy[1]
    assert original.ts.frame == 0
    assert copy.ts.frame == 1


def test_reader_initial_frame_maintained(original_and_copy):
    # check that copy inherits nonzero frame of original
    original, _ = original_and_copy

    if len(original) < 2:
        pytest.skip('Single frame reader')

    # seek
    original[1]

    copy = original.copy()

    assert original.ts.frame == copy.ts.frame
    assert_equal(original.ts.positions, copy.ts.positions)

def test_reader_initial_next(original_and_copy):
    # check that the filehandle (or equivalent) in the copied Reader
    # is identical to the original
    # ie calling next on both should produce identical results
    original, _ = original_and_copy

    if len(original) < 3:
        pytest.skip('Requires 3 frames')

    original[1]

    copy = original.copy()

    assert original.ts.frame == copy.ts.frame
    assert_equal(original.ts.positions, copy.ts.positions)

    original.next()
    copy.next()
    
    assert original.ts.frame == copy.ts.frame
    assert_equal(original.ts.positions, copy.ts.positions)


def test_timestep_copied(ref_reader):
    # modify the positions and dimensions from original
    # then check that the copy gets this (even though not in file)
    ref_reader.ts.positions *= 2
    ref_reader.ts.dimensions = newbox = [4, 5, 6, 70, 80, 90]
    new = ref_reader.copy()

    assert_equal(ref_reader.ts.positions, new.ts.positions)
    assert_almost_equal(new.ts.dimensions, newbox, decimal=4)
    assert ref_reader.ts.positions.dtype == np.float32
    assert new.ts.positions.dtype == np.float32


@pytest.mark.skipif(shares_memory == False,
                    reason='old numpy lacked shares_memory')
def test_positions_share_memory(original_and_copy):
    # check that the memory in Timestep objects is unique
    original, copy = original_and_copy
    assert not np.shares_memory(original.ts.positions, copy.ts.positions)

    original.ts.positions *= 2

    with pytest.raises(AssertionError):
        assert_equal(original.ts.positions, copy.ts.positions)


def test_auxiliary_NIE():
    # Aux's not implemented, check for sane error message
    u = mda.Universe(XYZ_mini)

    u.trajectory.add_auxiliary('myaux', AUX_XVG)

    with pytest.raises(NotImplementedError) as e:
        u.trajectory.copy()
    assert 'Copy not implemented for AuxReader' in str(e.value)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2018 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

from numpy.testing import assert_almost_equal

import MDAnalysis as mda
from MDAnalysis.analysis.rdf import InterRDF_s, InterRDF

from MDAnalysisTests.datafiles import GRO_MEMPROT, XTC_MEMPROT


@pytest.fixture(scope='module')
def u():
    return mda.Universe(GRO_MEMPROT, XTC_MEMPROT)


@pytest.fixture(scope='module')
def sels(u):
    s1 = u.select_atoms('name ZND and resid 289')
    s2 = u.select_atoms(
         '(name OD1 or name OD2) and resid 51 and sphzone 5.0 (resid 289)')
    s3 = u.select_atoms('name ZND and (resid 291 or resid 292)')
    s4 = u.select_atoms('(name OD1 or name OD2) and sphzone 5.0 (resid 291)')
    ags = [[s1, s2], [s3, s4]]
    return ags


@pytest.fixture(scope='module')
def rdf(u, sels):
    return InterRDF_s(u, sels).run()


def test_nbins(u, sels):
    rdf = InterRDF_s(u, sels, nbins=412).run()

    assert len(rdf.results.bins) == 412


def test_range(u, sels):
    rmin, rmax = 1.0, 13.0
    rdf = InterRDF_s(u, sels, range=(rmin, rmax)).run()

    assert rdf.results.edges[0] == rmin
    assert rdf.results.edges[-1] == rmax


def test_count_size(rdf):
    # ZND vs OD1 & OD2
    # should see 2 elements in rdf.results.count
    # 1 element in rdf.results.count[0]
    # 2 elements in rdf.results.count[0][0]
    # 2 elements in rdf.results.count[1]
    # 2 elements in rdf.results.count[1][0]
    # 2 elements in rdf.results.count[1][1]
    assert len(rdf.results.count) == 2
    assert len(rdf.results.count[0]) == 1
    assert len(rdf.results.count[0][0]) == 2
    assert len(rdf.results.count[1]) == 2
    assert len(rdf.results.count[1][0]) == 2
    assert len(rdf.results.count[1][1]) == 2


def test_count(rdf):
    # should see one distance with 5 counts in count[0][0][1]
    # should see one distance with 3 counts in count[1][1][0]
    sel0 = rdf.results.count[0][0][1] == 5
    sel1 = rdf.results.count[1][1][0] == 3
    assert len(rdf.results.count[0][0][1][sel0]) == 1
    assert len(rdf.results.count[1][1][0][sel1]) == 1


def test_double_run(rdf):
    # running rdf twice should give the same result
    sel0 = rdf.results.count[0][0][1] == 5
    sel1 = rdf.results.count[1][1][0] == 3
    assert len(rdf.results.count[0][0][1][sel0]) == 1
    assert len(rdf.results.count[1][1][0][sel1]) == 1


def test_cdf(rdf):
    rdf.get_cdf()
    ref = rdf.results.count[0][0][0].sum()/rdf.n_frames
    assert rdf.results.cdf[0][0][0][-1] == ref


@pytest.mark.parametrize("density, value", [
    (None, 26551.55088100731),    # default, like False (no kwarg, see below)
    (False, 26551.55088100731),
    (True, 0.021915460340071267)])
def test_density(u, sels, density, value):
    kwargs = {'density': density} if density is not None else {}
    rdf = InterRDF_s(u, sels, **kwargs).run()
    assert_almost_equal(max(rdf.results.rdf[0][0][0]), value)
    if not density:
        s1 = u.select_atoms('name ZND and resid 289')
        s2 = u.select_atoms(
                'name OD1 and resid 51 and sphzone 5.0 (resid 289)')
        rdf_ref = InterRDF(s1, s2).run()
        assert_almost_equal(rdf_ref.results.rdf, rdf.results.rdf[0][0][0])


@pytest.mark.parametrize("attr", ("rdf", "bins", "edges", "count", "cdf"))
def test_rdf_attr_warning(rdf, attr):
    if attr == "cdf":
        rdf.get_cdf()
    wmsg = f"The `{attr}` attribute was deprecated in MDAnalysis 2.0.0"
    with pytest.warns(DeprecationWarning, match=wmsg):
        getattr(rdf, attr) is rdf.results[attr]



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis
from MDAnalysis.analysis import waterdynamics
from MDAnalysis.lib.correlations import autocorrelation, correct_intermittency

from MDAnalysisTests.datafiles import waterPSF, waterDCD

import pytest
import numpy as np
from unittest.mock import patch, Mock
from numpy.testing import assert_almost_equal, assert_equal

SELECTION1 = "byres name OH2"
SELECTION2 = "byres name P1"


@pytest.fixture(scope='module')
def universe():
    return MDAnalysis.Universe(waterPSF, waterDCD)


def test_WaterOrientationalRelaxation(universe):
    wor = waterdynamics.WaterOrientationalRelaxation(
        universe, SELECTION1, 0, 5, 2)
    wor.run()
    assert_almost_equal(wor.timeseries[1][2], 0.35887,
                        decimal=5)


def test_WaterOrientationalRelaxation_zeroMolecules(universe):
    wor_zero = waterdynamics.WaterOrientationalRelaxation(
        universe, SELECTION2, 0, 5, 2)
    wor_zero.run()
    assert_almost_equal(wor_zero.timeseries[1], (0.0, 0.0, 0.0))


def test_AngularDistribution(universe):
    ad = waterdynamics.AngularDistribution(universe, SELECTION1, 40)
    ad.run()
    # convert a string with two "floats" into a float array
    result = np.array(ad.graph[0][39].split(), dtype=np.float64)
    assert_almost_equal(result, (0.951172947884, 0.48313682125))


def test_MeanSquareDisplacement(universe):
    msd = waterdynamics.MeanSquareDisplacement(universe, SELECTION1, 0, 10, 2)
    msd.run()
    assert_almost_equal(msd.timeseries[1], 0.03984,
                        decimal=5)


def test_MeanSquareDisplacement_zeroMolecules(universe):
    msd_zero = waterdynamics.MeanSquareDisplacement(
        universe, SELECTION2, 0, 10, 2)
    msd_zero.run()
    assert_almost_equal(msd_zero.timeseries[1], 0.0)


def test_SurvivalProbability_intermittency1and2(universe):
    """
    Intermittency of 2 means that we still count an atom if it is not present for up to 2 consecutive frames,
    but then returns at the following step.
    """
    with patch.object(universe, 'select_atoms') as select_atoms_mock:
        ids = [(9, 8), (), (8,), (9,), (8,), (), (9, 8), (), (8,), (9, 8)]
        select_atoms_mock.side_effect = lambda selection: Mock(ids=ids.pop())   # atom IDs fed set by set
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=3, stop=10, verbose=True, intermittency=2)
        assert all(x == {9, 8} for x in sp._intermittent_selected_ids)
        assert_almost_equal(sp.sp_timeseries, [1, 1, 1, 1])


def test_SurvivalProbability_intermittency2lacking(universe):
    """
    If an atom is not present for more than 2 consecutive frames,
    it is considered to have left the region.
    """
    with patch.object(universe, 'select_atoms') as select_atoms_mock:
        ids = [(9,), (), (), (), (9,), (), (), (), (9,)]
        select_atoms_mock.side_effect = lambda selection: Mock(ids=ids.pop())   # atom IDs fed set by set
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=3, stop=9, verbose=True, intermittency=2)
        assert_almost_equal(sp.sp_timeseries, [1, 0, 0, 0])


def test_SurvivalProbability_intermittency1_step5_noSkipping(universe):
    """
    Step leads to skipping frames if (tau_max + 1) + (intermittency * 2) < step.
    No frames should be skipped.
    """
    with patch.object(universe, 'select_atoms') as select_atoms_mock:
        ids = [(2, 3), (3,), (2, 3), (3,), (2,), (3,), (2, 3), (3,), (2, 3), (2, 3)]
        select_atoms_mock.side_effect = lambda selection: Mock(ids=ids.pop())   # atom IDs fed set by set
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=2, stop=10, verbose=True, intermittency=1, step=5)
        assert all((x == {2, 3} for x in sp._intermittent_selected_ids))
        assert_almost_equal(sp.sp_timeseries, [1, 1, 1])


def test_SurvivalProbability_intermittency1_step5_Skipping(universe):
    """
    Step leads to skipping frames if (tau_max + 1) * (intermittency * 2) < step.
    In this case one frame will be skipped per window.
    """
    with patch.object(universe, 'select_atoms') as select_atoms_mock:
        ids = [(1,), (), (1,), (), (1,), (), (1,), (), (1,), (1,)]
        beforepopsing = len(ids) - 2
        select_atoms_mock.side_effect = lambda selection: Mock(ids=ids.pop())   # atom IDs fed set by set
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=1, stop=10, verbose=True, intermittency=1, step=5)
        assert all((x == {1} for x in sp._intermittent_selected_ids))
        assert len(sp._selected_ids) == beforepopsing
        assert_almost_equal(sp.sp_timeseries, [1, 1])


def test_intermittency_none():
    # No changes asked - returns the same data
    input_ids = [{1}, {1}, {1}, set(), set(), {1}, {1}, {1}, set(), set(), {1}, set(), set(), {1}]
    corrected = correct_intermittency(input_ids, intermittency=0)
    assert all(x == y for x,y in zip(input_ids, corrected))


def test_intermittency_1and2():
    # The maximum gap in the dataset is 2, so the IDs are always present after correction
    input_ids = [{9, 8}, set(), {8, }, {9, }, {8, }, set(), {9, 8}, set(), {8, }, {9, 8, }]
    corrected = correct_intermittency(input_ids, intermittency=2)
    assert all((x == {9, 8} for x in corrected))


def test_intermittency_2tooShort():
    #The IDs are abscent for too long/
    input_ids = [{9,}, {}, {}, {}, {9,}, {}, {}, {}, {9,}]
    corrected = correct_intermittency(input_ids, intermittency=2)
    assert all(x == y for x, y in zip(input_ids, corrected))


def test_intermittency_setsOfSets():
    # Verificaiton for the case of hydrogen bonds (sets of sets)
    input_ids = [{frozenset({1,2}), frozenset({3, 4})},set(), set(),
                 {frozenset({1, 2}), frozenset({3, 4})}, set(), set(),
                 {frozenset({1, 2}), frozenset({3, 4})}, set(), set(),
                 {frozenset({1, 2}), frozenset({3, 4})}]
    corrected = correct_intermittency(input_ids, intermittency=2)
    assert all((x == {frozenset({1, 2}), frozenset({3, 4})} for x in corrected))


def test_autocorrelation_alwaysPresent():
    input = [{1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}]
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(input, tau_max=3)
    assert all(np.equal(sp_timeseries, 1))


def test_autocorrelation_definedTaus():
    input_ids = [{9, 8, 7}, {8, 7, 6}, {7, 6, 5}, {6, 5, 4}, {5, 4, 3}, {4, 3, 2}, {3, 2, 1}]
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(input_ids, tau_max=3)
    assert_almost_equal(sp_timeseries, [1, 2/3., 1/3., 0])


def test_autocorrelation_intermittency1_windowJump_intermittencyAll():
    """
    Step leads to skipping frames if (tau_max + 1) + (intermittency * 2) < step.
    No frames should be skipped so intermittency should be applied to all.
    """
    input_ids = [{2, 3}, {3,}, {2, 3}, {3,}, {2,}, {3,}, {2, 3}, {3,}, {2, 3}, {2, 3}]
    corrected = correct_intermittency(input_ids, intermittency=1)
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(corrected, tau_max=2,
                                                                        window_step=5)
    assert all((x == {2, 3} for x in corrected))
    assert_almost_equal(sp_timeseries, [1, 1, 1])


def test_autocorrelation_windowBigJump():
    #The empty sets are ignored (no intermittency)
    input_ids = [{1}, {1}, {1}, set(), set(), {1}, {1}, {1}, set(), set(), {1}, {1}, {1}]
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(input_ids, tau_max=2, window_step=5)
    assert_almost_equal(sp_timeseries, [1, 1, 1])


def test_autocorrelation_windowBigJump_absence():
    # In the last frame the molecules are absent
    input_ids = [{1}, {1}, {1}, set(), set(), {1}, {1}, {1}, set(), set(), {1}, set(), set()]
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(input_ids, tau_max=2, window_step=5)
    assert_almost_equal(sp_timeseries, [1, 2/3., 2/3.])


def test_autocorrelation_intermittency1_many():
    input_ids = [{1}, set(), {1}, set(), {1}, set(), {1}, set(), {1}, set(), {1}, set(), {1}, set(), {1}]
    corrected = correct_intermittency(input_ids, intermittency=1)
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(corrected, tau_max=14,
                                                                        window_step=5)
    assert_almost_equal(sp_timeseries, [1] * 15)


def test_autocorrelation_intermittency2_windowBigJump():
    # The intermittency corrects the last frame
    input_ids = [{1}, {1}, {1}, set(), set(), {1}, {1}, {1}, set(), set(), {1}, set(), set(), {1}]
    corrected = correct_intermittency(input_ids, intermittency=2)
    tau_timeseries, sp_timeseries, sp_timeseries_data = autocorrelation(corrected, tau_max=2,
                                                                        window_step=5)
    assert_almost_equal(sp_timeseries, [1, 1, 1])


def test_SurvivalProbability_t0tf(universe):
    with patch.object(universe, 'select_atoms') as select_atoms_mock:
        ids = [(0, ), (0, ), (7, 6, 5), (6, 5, 4), (5, 4, 3), (4, 3, 2), (3, 2, 1), (0, )]
        select_atoms_mock.side_effect = lambda selection: Mock(ids=ids.pop(2))   # atom IDs fed set by set
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=3, start=2, stop=7)
        assert_almost_equal(sp.sp_timeseries, [1, 2 / 3.0, 1 / 3.0, 0])


def test_SurvivalProbability_definedTaus(universe):
    with patch.object(universe, 'select_atoms') as select_atoms_mock:
        ids = [(9, 8, 7), (8, 7, 6), (7, 6, 5), (6, 5, 4), (5, 4, 3), (4, 3, 2), (3, 2, 1)]
        select_atoms_mock.side_effect = lambda selection: Mock(ids=ids.pop())   # atom IDs fed set by set
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=3, start=0, stop=7, verbose=True)
        assert_almost_equal(sp.sp_timeseries, [1, 2 / 3.0, 1 / 3.0, 0])


def test_SurvivalProbability_zeroMolecules(universe):
    # no atom IDs found
    with patch.object(universe, 'select_atoms', return_value=Mock(ids=[])) as select_atoms_mock:
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=3, start=3, stop=7, verbose=True)
        assert all(np.isnan(sp.sp_timeseries[1:]))


def test_SurvivalProbability_alwaysPresent(universe):
    # always the same atom IDs found, 7 and 8
    with patch.object(universe, 'select_atoms', return_value=Mock(ids=[7, 8])) as select_atoms_mock:
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=3, start=0, stop=7, verbose=True)
        assert all(np.equal(sp.sp_timeseries, 1))


def test_SurvivalProbability_stepLargerThanDtmax(universe):
    # Testing if the frames are skipped correctly
    with patch.object(universe, 'select_atoms', return_value=Mock(ids=(1,))) as select_atoms_mock:
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=2, step=5, stop=10, verbose=True)
        assert_equal(sp.sp_timeseries, [1, 1, 1])
        # with tau_max=2 for all the frames we only read 6 of them
        # this is because the frames which are not used are skipped, and therefore 'select_atoms'
        assert universe.trajectory.n_frames > 6
        assert_equal(select_atoms_mock.call_count, 6)


def test_SurvivalProbability_stepEqualDtMax(universe):
    with patch.object(universe, 'select_atoms', return_value=Mock(ids=(1,))) as select_atoms_mock:
        sp = waterdynamics.SurvivalProbability(universe, "")
        sp.run(tau_max=4, step=5, stop=10, verbose=True)
        # all frames from 0, with 9 inclusive
        assert_equal(select_atoms_mock.call_count, 10)
